/********************************************************************\
 * Mu (Miscellaneous Utilities) is a general convenience library.   *
 * Copyright (C) 2020  Asher Gordon <AsDaGo@posteo.net>             *
 *                                                                  *
 * This file is part of Mu.                                         *
 *                                                                  *
 * Mu is free software: you can redistribute it and/or modify it    *
 * under the terms of the GNU General Public License as published   *
 * by the Free Software Foundation, either version 3 of the         *
 * License, or (at your option) any later version.                  *
 *                                                                  *
 * Mu is distributed in the hope that it will be useful, but        *
 * WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU General Public License for more details.                     *
 *                                                                  *
 * You should have received a copy of the GNU General Public        *
 * License along with Mu.  If not, see                              *
 * <https://www.gnu.org/licenses/>.                                 *
\********************************************************************/

/*
 * options.c -- routines for parsing options
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#if defined HAVE_WORKING_FORK && \
  defined HAVE_DUP2 && defined HAVE_FCNTL_H
# define USE_MAN_PIPE 1
#endif

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <libgen.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <signal.h>
#include <dirent.h>
#include <limits.h>
#include <errno.h>
#include <assert.h>
#ifdef USE_MAN_PIPE
# include <fcntl.h>
#endif

#include "sig2str.h"
#include "safe.h"
#include "format.h"
#include "compat.h"
#include "error.h"
#include "options.h"

#define MANDATORY_ARGS_MSG			\
  "Mandatory arguments to long options "	\
  "are mandatory for short options too."

#define MAN_PARAGRAPH_WIDTH 70

#define OPT_END(opt)					\
  ((opt)->short_opt == '\0' &&				\
   (opt)->long_opt == NULL && (opt)->env_var == NULL)
#define SUBOPT_END(opt) ((opt)->subopt_name == NULL)
#define ANYOPT_END(opt, is_subopt)			\
  ((is_subopt) ? SUBOPT_END(opt) : OPT_END(opt))

/* The argument separator for long options, suboptions, and
   environment variables. */
#define ARG_SEPARATOR '='
/* The separator for suboptions. */
#define SUBOPT_SEPARATOR ','

/* The absolute maximum line length for usage messages. Undefine for
   no max. */
#define USAGE_MAX_WIDTH		80

/* The default line length for usage messages. */
#define USAGE_DEFAULT_WIDTH	USAGE_MAX_WIDTH

/* Types to pass to handle_option(). */
enum opt_type {
  OPT_TYPE_SHORT,
  OPT_TYPE_LONG,
  OPT_TYPE_ENV,
  OPT_TYPE_SUBOPT
};

/* The format in which to output help text. */
enum help_type {
  HELP_TYPE_PLAIN,
  HELP_TYPE_MAN
};

typedef union opt_arg {
  int arg_bool;
  long arg_int;
  double arg_float;
  const char *arg_string;
  FILE *arg_file;
  DIR *arg_directory;
} OPT_ARG;

/* A callback to pass to the help formatting functions. It should
   return zero on success or nonzero on error, in which case `errno'
   should be set to indicate the error. */
typedef int FORMAT_HELP_CALLBACK(void *, unsigned short,
				 unsigned short, unsigned short *,
				 unsigned short, unsigned short,
				 const char *, ...)
/* Yep, we can put attributes on function typedefs. */
  __attribute__((format(printf, 7, 8)));

/* A list of strings. */
struct string_list {
  char *str;
  struct string_list *next;
};

/* This is for formatting the usage message with
   help_option_callback(). */
struct help_callback_data {
  unsigned ref_count;
  const char *default_format, *prog_name, *usage, *short_desc,
    *desc, *notes, *section, *sect_name, *pkg_str, *date;
  MU_OPT *options;
};

/*
 * Static helper functions
 */
static void switch_args(char **, int, int);
static size_t get_opts_len(const MU_OPT *);
static void init_opts(const char *, const MU_OPT *, int);
static int parse_opt(const char *, char *, char *, const MU_OPT *, int);
static int parse_short_opt(const char *, char *, char *,
			   const MU_OPT *, int, int);
static int parse_long_opt(const char *, const char *,
			  char *, const MU_OPT *, int, int);
static int parse_env_var(const char *, const char *,
			 const MU_OPT *, int);
static int check_env_var(const char *, const MU_OPT *,
			 const char *, size_t);
static int handle_option(const char *, enum opt_type,
			 const MU_OPT *, char *, ...);
static unsigned find_unambiguously(unsigned **, int *,
				   const MU_OPT *, const char *, int);
static int index_sort(const void *, const void *);
static const char *get_arg_help(const MU_OPT *) __attribute__((pure));
static const char *get_arg_separator(const MU_OPT *, enum opt_type)
  __attribute__((pure));
static size_t get_min_indent(const MU_OPT *, enum opt_type, int)
  __attribute__((pure));
static int free_opt_arg(const char *, const MU_OPT *);
static int free_opt_args(const char *, const MU_OPT *);
static int format_help_plain(const char *, char **, const char *,
			     const char *, const MU_OPT *,
			     FORMAT_HELP_CALLBACK, void *,
			     unsigned short, unsigned short);
static int format_help_man(const char *, char **, const char *,
			   const char *, const char *, const char *,
			   const char *, const char *, const char *,
			   const MU_OPT *,
			   FORMAT_HELP_CALLBACK, void *);
static char **create_usage_list(const char *);
static void free_usage_list(char **);
static int print_subopts_header(const MU_OPT *, const char *,
				const char *, const char *,
				FORMAT_HELP_CALLBACK, void *,
				unsigned short *,
				unsigned short, unsigned short);
static int format_opt_help(const MU_OPT *, enum opt_type, int,
			   FORMAT_HELP_CALLBACK, void *,
			   unsigned short, unsigned short,
			   unsigned short *,
			   unsigned short, unsigned short);
static int format_opt_help_man(const MU_OPT *, enum opt_type,
			       FORMAT_HELP_CALLBACK, void *,
			       unsigned short *);
static const char *get_opt_help(const char *, enum opt_type)
  __attribute__((pure));
static char *format_man_str(const char *, int)
  __attribute__((malloc));
static size_t format_man_str_no_alloc(char *, size_t,
				      const char *, int);
static int write_out(char *, size_t *, size_t, const char *, ...)
  __attribute__((format(printf, 4, 5)));
static int is_var_char(int);
static int is_opt_char(int);
static int is_word_char(int);
static FORMAT_HELP_CALLBACK format_help_file_callback;
static FORMAT_HELP_CALLBACK format_help_str_callback;
static char *concat_string_list(struct string_list *,
				struct string_list *)
  __attribute__((malloc));
static void free_string_list(struct string_list *,
			     struct string_list *);
static int help_option_callback(int, const char *, void *, char *);
static int man_option_callback(void *, char *);
static void free_help_callback_data(void *);
#ifdef USE_MAN_PIPE
static void report_err(int, const char *, ...)
  __attribute__((format(printf, 2, 3), noreturn));
static char *get_err(int) __attribute__((malloc));
#endif

/* Shift arguments while keeping argv[0] in place. */
void mu_shift_args(int *argc, char ***argv, int amount) {
  *argc -= amount;
  *argv += amount;
  (*argv)[0] = (*argv)[-amount];
}

/*
 * Parse options. If MU_OPT_PERMUTE is passed in `flags' and the
 * environment variable POSIXLY_CORRECT is not set, '*argv' + the
 * return value will contain only non-option positional arguments
 * (plus 'argv[0]') when we're done parsing. Otherwise, if
 * MU_OPT_PERMUTE is not passed in `flags' or the environment variable
 * POSIXLY_CORRECT is set, we will stop parsing at the first
 * non-option argument and '*argv' will be left alone. If
 * `MU_OPT_CONTINUE' is passed in `flags', an additional argument
 * specifies a pointer to an integer which tells us whether we should
 * stop parsing options (should be initialized to 0).
 *
 * `env' is the environment used for parsing environment variables. If
 * it is passed as NULL, no environment variable parsing will be
 * performed.
 *
 * Returns the number of arguments parsed (options and arguments to
 * those options). on success. You should pass this value to
 * mu_shift_args(). On error, an error code is returned and an error
 * message will be printed to `stderr'. The last element of `options'
 * must have a `short_opt' of '\0', a `long_opt' of NULL, and an
 * `env_var' of NULL.
 */
int mu_parse_opts(int argc, char **argv, char **env,
		  MU_OPT *options, int flags, ...) {
  int ret = 0, nonopt_index = 1, args_parsed = 0;
  MU_OPT_DATA *data;

  if (flags & MU_OPT_CONTINUE) {
    va_list ap;

    va_start(ap, flags);
    data = va_arg(ap, MU_OPT_DATA *);
    va_end(ap);

    assert(data);

    if (data->stop_parsing)
      return 0;
  }
  else {
    /* Stifle -Wmaybe-uninitialized (GCC is smart enough to realize
       that 'flags &= ~MU_OPT_PERMUTE' might cause 'flags &
       MU_OPT_CONTINUE' to be 0, but it's not smart enough to realize
       that it actually won't). */
    data = NULL;
  }

  init_opts(argv[0], options, 0);

  /* Remove MU_OPT_PERMUTE from `flags' if the environment variable
     POSIXLY_CORRECT is set. */
  if (flags & MU_OPT_PERMUTE && getenv("POSIXLY_CORRECT"))
    flags &= ~MU_OPT_PERMUTE;

  /* Parse the environment first (because command line options should
     take precedence). */
  if (env && !(flags & MU_OPT_CONTINUE && data->skip_env)) {
    for (; *env; env++) {
      ret = parse_env_var(argv[0], *env, options, flags);
      if (MU_OPT_ERR(ret)) {
	/* Error! */
	free_opt_args(argv[0], options);
	args_parsed = ret;
	goto done;
      }
    }
  }

  /* Parse the command line options. */
  for (int i = 1; i < argc; i++) {
    char *optstr = argv[i];
    char *arg = (i + 1 < argc) ? argv[i + 1] : NULL;

    /* Stop if we see "--". */
    if (!strcmp(optstr, "--")) {
      /* Move "--" back with the other options, but only if
	 MU_OPT_ALLOW_INVALID is *not* passed in `flags'. This is
	 because if MU_OPT_ALLOW_INVALID is passed, we are probably
	 going to get called again with a different set of
	 options. And when we are, we want to see this "--" again so
	 we don't continue parsing options. */
      if (!(flags & MU_OPT_ALLOW_INVALID))
	switch_args(argv, nonopt_index++, i);
      args_parsed++;

      /* Remember to stop parsing options if we're going to be called
	 again. */
      if (flags & MU_OPT_CONTINUE)
	data->stop_parsing = 1;

      break;
    }

    /* Parse the option. */
    ret = parse_opt(argv[0], optstr, arg, options, flags);

    if (ret == 2) {
      /* It's not an option. */
      if (!(flags & MU_OPT_PERMUTE))
	break;
    }
    else
      args_parsed++;

    if (MU_OPT_ERR(ret)) {
      /* Error! */
      free_opt_args(argv[0], options);
      args_parsed = ret;
      break;
    }

    if (ret != 2 && flags & MU_OPT_PERMUTE)
      switch_args(argv, nonopt_index++, i);

    if (ret == 1) {
      /* For the argument to the option. */
      i++;
      args_parsed++;
      if (flags & MU_OPT_PERMUTE)
	switch_args(argv, nonopt_index++, i);
    }
  }

 done:
  /* Free the `data' field of all options that we need to but only if
     we're not going to be run again. */
  if (!(flags & MU_OPT_CONTINUE))
    mu_free_opts(options);

  return args_parsed;
}

/*
 * Parse suboptions. We export this function in case the user has some
 * strange option parsing needs (for example, suboption-style options
 * in a positional argument or something). `prog_name' is used for
 * reporting errors. `arg' must be writable, because we overwrite
 * commas in `arg'. Returns 0 on success, or an error code on error.
 */
int mu_parse_subopts(const char *prog_name,
		     char *suboptstr, MU_OPT *subopts) {
  int ret = 0;
  char *next;
  const MU_OPT *subopt;

  init_opts(prog_name, subopts, 1);

  for (char *str = suboptstr; str; str = next) {
    unsigned *matches, n_matches;
    char *argstr;

    next = strchr(str, SUBOPT_SEPARATOR);
    if (next)
      *(next++) = '\0';

    /* A suboption cannot start with ARG_SEPARATOR. */
    if (*str == ARG_SEPARATOR) {
      fprintf(stderr, "%s: '%s': invalid suboption\n", prog_name, str);
      ret = MU_OPT_ERR_PARSE;
      break;
    }

    /* Search for the suboption. */
    n_matches = find_unambiguously(&matches, NULL, subopts, str, 1);
    if (n_matches != 1) {
      if (n_matches) {
	fprintf(stderr,
		"%s: '%s': suboption is ambiguous: possibilities:\n",
		prog_name, suboptstr);
	/* Sort the matches so they appear in the same order as they
	   do in the help message. */
	qsort(matches, n_matches, sizeof(*matches), index_sort);
	for (unsigned i = 0; i < n_matches; i++)
	  fprintf(stderr, "  %s\n", subopts[matches[i]].subopt_name);
      }
      else
	fprintf(stderr, "%s: '%s': invalid suboption\n",
		prog_name, suboptstr);

      free(matches);

      ret = MU_OPT_ERR_PARSE;
      break;
    }

    subopt = subopts + *matches;

    free(matches);

    argstr = strchr(str, ARG_SEPARATOR);
    if (argstr)
      argstr++;

    ret = handle_option(prog_name, OPT_TYPE_SUBOPT, subopt, argstr);
    if (ret)
      break;
  }

  if (ret) {
    /* An error occurred; clean up. */
    for (subopt = subopts; !SUBOPT_END(subopt); subopt++)
      free_opt_arg(prog_name, subopt);
  }

  return ret;
}

/*
 * Free all callback data that needs to be freed.
 */
void mu_free_opts(MU_OPT *options) {
  for (MU_OPT *opt = options; !OPT_END(opt); opt++) {
    if (opt->has_arg && opt->arg_type == MU_OPT_SUBOPT)
      mu_free_opts(opt->subopts);
    else if (opt->cb_data && opt->cb_data_destructor)
      opt->cb_data_destructor(opt->cb_data);
  }
}

/*
 * Concatenate option lists `a' and `b', returning the result in
 * '*res'. If either one is NULL, it will be treated as empty. If
 * `extra' is nonzero, extra space will be allocated for `extra' more
 * options. Returns the size of '*res' on success, or 0 on error.
 */
size_t mu_concat_opts(MU_OPT **res, const MU_OPT *a,
		      const MU_OPT *b, size_t extra) {
  size_t a_len, b_len, res_len, a_size, b_size, res_size;
  const MU_OPT empty[] = { 0 };

  if (!a)
    a = empty;
  if (!b)
    b = empty;

  /* Get the lengths and sizes of `a' and `b'. */
  a_len = get_opts_len(a);
  b_len = get_opts_len(b);
  a_size = a_len * sizeof(*a);
  b_size = b_len * sizeof(*b);

  /* Get the length and size of '*res'. */
  res_len = a_len + b_len + extra + 1; /* +1 for terminating option */
  res_size = res_len * sizeof(**res);

  /* Concatenate the options. */
  *res = malloc(res_size);
  if (!*res)
    return 0;
  memcpy(*res,		a, a_size);
  memcpy(*res + a_len,	b, b_size);

  /* Initialize the extra options and the terminating option to
     zero. */
  memset(*res + a_len + b_len, 0, res_size - a_size - b_size);

  return res_len;
}

/*
 * Add a help option to `options'. If MU_HELP_SHORT is passed in
 * `flags', the '-h' option will be added and if MU_HELP_LONG is
 * passed in `flags', the '--help' option will be added. MU_HELP_BOTH
 * is a synonym for 'MU_HELP_SHORT | MU_HELP_LONG'. The help option
 * will be added to the first option that has a `short_opt' of '\0'
 * and a `long_opt' of NULL).
 */
void mu_create_help_option(MU_OPT *options, const char *prog_name,
			   const char *usage, const char *short_desc,
			   const char *desc, const char *notes,
			   const char *section, const char *sect_name,
			   const char *pkg_str, const char *date,
			   int flags) {
  struct help_callback_data *data;
  MU_OPT *opt;

  /* Make sure we're creating at least one option. */
  assert(flags & (MU_HELP_BOTH | MU_HELP_MAN_BOTH));

  /* Better to find out sooner rather than later (in mu_format_help()
     or mu_format_help_man()). */
  assert(prog_name);

  data = mu_xmalloc(sizeof(*data));

  data->ref_count	= 0;
  data->default_format	= NULL;
  data->prog_name	= prog_name;
  data->usage		= usage;
  data->short_desc	= short_desc;
  data->desc		= desc;
  data->notes		= notes;
  data->section		= section;
  data->sect_name	= sect_name;
  data->pkg_str		= pkg_str;
  data->date		= date;
  data->options		= options;

  if (flags & MU_HELP_BOTH) {
#define HELP_START "\
Print this help and exit. If FORMAT is \"man\", format the help as a \
man(1) page. If FORMAT is \"plain\" or \"text\", format the help as \
plain text. If FORMAT is omitted, it will default to"
#define HELP_END "\"plain\"."
    const char *const help_no_env = HELP_START " " HELP_END;
    const char *const help_env = HELP_START " the value of the "
      "MU_HELP_FORMAT environment variable if set, otherwise "
      HELP_END;
#undef HELP_START
#undef HELP_END
    const char *const help =
      (flags & MU_HELP_ENV) ? help_env : help_no_env;

    /* Go to the option which we want to make into a help option. */
    for (opt = options; !OPT_END(opt); opt++);

    memset(opt, 0, sizeof(*opt));

    if (flags & MU_HELP_SHORT)
      opt->short_opt		= 'h';
    if (flags & MU_HELP_LONG)
      opt->long_opt		= "help";
    opt->has_arg		= MU_OPT_OPTIONAL;
    opt->arg_type		= MU_OPT_STRING;
    opt->callback_string	= help_option_callback;
    opt->cb_data		= data;
    opt->cb_data_destructor	= free_help_callback_data;
    opt->arg_help		= "FORMAT";
    opt->help			= help;

    data->ref_count++;
  }
  else
    assert(!(flags & MU_HELP_ENV));

  if (flags & MU_HELP_MAN_BOTH) {
    /* Go to the option which we want to make into a man option. */
    for (opt = options; !OPT_END(opt); opt++);

    memset(opt, 0, sizeof(*opt));

    if (flags & MU_HELP_MAN_SHORT)
      opt->short_opt		= 'm';
    if (flags & MU_HELP_MAN_LONG)
      opt->long_opt		= "man";
    opt->has_arg		= MU_OPT_NONE;
    opt->callback_none		= man_option_callback;
    opt->cb_data		= data;
    opt->cb_data_destructor	= free_help_callback_data;
    opt->help			= "print this help as a man(1) page";

    data->ref_count++;
  }

  if (flags & MU_HELP_ENV) {
#define HELP_START "the default format for "
    const char *const help_short	= HELP_START "-h";
    const char *const help_long		= HELP_START "--help";
    const char *const help_both		= HELP_START "-h, --help";
#undef HELP_START
    const char *help;

    if (flags & MU_HELP_SHORT)
      help = (flags & MU_HELP_LONG) ? help_both : help_short;
    else
      help = help_long;

    /* Go to the option which we want to make into the MU_HELP_FORMAT
       environment variable. */
    for (opt = options; !OPT_END(opt); opt++);

    memset(opt, 0, sizeof(*opt));

    opt->env_var	= "MU_HELP_FORMAT";
    opt->has_arg	= MU_OPT_OPTIONAL;
    opt->arg_type	= MU_OPT_STRING;
    opt->arg		= &(data->default_format);
    opt->arg_help	= "FORMAT";
    opt->help		= help;
  }
}

/*
 * Format help output from `usage' and `options', printing it to
 * `file'. Returns zero on success, nonzero on error in which case
 * `errno' will be set to indicate the error.
 */
int mu_format_help(FILE *file, const char *prog_name,
		   const char *usage, const char *desc,
		   const char *notes, const MU_OPT *options) {
  unsigned short goal, width;
  char *columns_str;
  int columns;
  char **usage_list;
  int ret;
#ifdef TIOCGWINSZ
  struct winsize ws;
#endif

  /* COLUMNS should take precedence in case the user wants us to act
     as though we were on a different sized screen (e.g. for testing
     purposes). */
  columns_str = getenv("COLUMNS");
  columns = columns_str ? atoi(columns_str) : 0;

  if (columns < 1) {
    /* Either the COLUMNS environment variable was not set, or it was
       not a valid integer, or it was less than 1. In any case, get the
       terminal width if possible, otherwise using the default width. */
#ifdef TIOCGWINSZ
    if (ioctl(fileno(file), TIOCGWINSZ, &ws) >= 0)
      columns = ws.ws_col;
    else
#endif
      columns = USAGE_DEFAULT_WIDTH;
  }

  assert(columns > 0);
  if (columns > USHRT_MAX)
    columns = USHRT_MAX;

  goal = width = columns;

#ifdef USAGE_MAX_WIDTH
  /* The help message will look kind of weird if the goal width is
     more than USAGE_MAX_WIDTH. */
  if (goal > USAGE_MAX_WIDTH)
    goal = USAGE_MAX_WIDTH;
#endif

  /* Finally format the help message. */
  usage_list = create_usage_list(usage);
  if (!usage_list)
    return 1;
  ret = format_help_plain(prog_name, usage_list, desc, notes, options,
			  format_help_file_callback, file, goal, width);
  free_usage_list(usage_list);
  return ret;
}

/*
 * Format help output from `usage' and `options'. Returns the
 * formatted help (a dynamically allocated string which must be freed)
 * on success or NULL on error, in which case `errno' will be set to
 * indicate the error.
 */
char *mu_format_help_string(const char *prog_name, const char *usage,
			    const char *desc, const char *notes,
			    const MU_OPT *options,
			    unsigned short goal,
			    unsigned short width) {
  struct string_list *first, *last;
  char **usage_list;
  int ret;

  first = last = malloc(sizeof(*first));
  if (!first)
    return NULL;

  usage_list = create_usage_list(usage);
  if (!usage_list)
    return NULL;

  /* Format the help message. */
  ret =
    format_help_plain(prog_name, usage_list, desc, notes, options,
		      format_help_str_callback, &last, goal, width);
  free_usage_list(usage_list);
  if (ret) {
    free_string_list(first, last);
    return NULL;
  }

  return concat_string_list(first, last);
}

/*
 * Format help output in man(7) format from `usage' and `options',
 * printing it to `file'. Returns zero on success, nonzero on error in
 * which case `errno' will be set to indicate the error.
 */
int mu_format_help_man(FILE *file, const char *prog_name,
		       const char *usage, const char *short_desc,
		       const char *desc, const char *notes,
		       const char *section, const char *sect_name,
		       const char *pkg_str, const char *date,
		       const MU_OPT *options) {
  int ret;
  char **usage_list;

  usage_list = create_usage_list(usage);
  if (!usage_list)
    return 1;
  ret = format_help_man(prog_name, usage_list, short_desc, desc,
			notes, section, sect_name, pkg_str, date,
			options, format_help_file_callback, file);
  free_usage_list(usage_list);
  return ret;
}

/*
 * Format help output in man(7) format from `usage' and
 * `options'. Returns the formatted help (a dynamically allocated
 * string which must be freed) on success or NULL on error, in which
 * case `errno' will be set to indicate the error.
 */
char *mu_format_help_man_string(const char *prog_name,
				const char *usage,
				const char *short_desc,
				const char *desc, const char *notes,
				const char *section,
				const char *sect_name,
				const char *pkg_str, const char *date,
				const MU_OPT *options) {
  struct string_list *first, *last;
  char **usage_list;
  int ret;

  first = last = malloc(sizeof(*first));
  if (!first)
    return NULL;

  usage_list = create_usage_list(usage);
  if (!usage_list)
    return NULL;

  /* Format the help message. */
  ret = format_help_man(prog_name, usage_list, short_desc, desc,
			notes, section, sect_name, pkg_str, date,
			options, format_help_str_callback, &last);
  free_usage_list(usage_list);
  if (ret) {
    free_string_list(first, last);
    return NULL;
  }

  return concat_string_list(first, last);
}

/***************************\
|* Static helper functions *|
\***************************/

/*
 * Set 'argv[a]' to 'argv[b]' and then shift the arguments over.
 */
static void switch_args(char **argv, int a, int b) {
  char *old_arg;

  assert(a >= 0);
  assert(b >= 0);

  if (a > b) {
    int c;
    c = a;
    a = b;
    b = c;
  }

  old_arg   = argv[a];
  argv[a++] = argv[b];
  while (a <= b) {
    char *tmp = old_arg;
    old_arg = argv[a];
    argv[a++] = tmp;
  }
}

/*
 * Return the length of a list of options (not including the
 * terminating empty option).
 */
static size_t get_opts_len(const MU_OPT *opts) {
  size_t len;
  for (len = 0; !OPT_END(opts + len); len++);
  return len;
}

/*
 * Initalize all the `found_opt', `found_arg', and `argstr' fields to
 * zero and the `arg' fields to the default values. Also validate the
 * options.
 */
static void init_opts(const char *prog_name,
		      const MU_OPT *options, int subopts) {
  for (const MU_OPT *opt = options; !ANYOPT_END(opt, subopts); opt++) {
    if (opt->long_opt && !*opt->long_opt)
      mu_die(-1, "%s: \"\" is not a valid long option name\n",
	     prog_name);

    /* Make sure there are no duplicates of this option. */
    for (const MU_OPT *check_opt = opt + 1;
	 !ANYOPT_END(check_opt, subopts); check_opt++) {
      if (subopts) {
	if (!strcmp(opt->subopt_name, check_opt->subopt_name))
	  mu_die(-1, "%s: duplicate suboption: '%s'\n",
		 prog_name, opt->subopt_name);
      }
      else {
	if (opt->short_opt && opt->short_opt == check_opt->short_opt)
	  mu_die(-1, "%s: duplicate option: '-%c'\n",
		 prog_name, opt->short_opt);
	if (opt->long_opt && check_opt->long_opt &&
	    !strcmp(opt->long_opt, check_opt->long_opt))
	  mu_die(-1, "%s: duplicate option: '--%s'\n",
		 prog_name, opt->long_opt);
	if (opt->env_var && check_opt->env_var &&
	    !strcmp(opt->env_var, check_opt->env_var))
	  mu_die(-1, "%s: duplicate environment variable: '%s'\n",
		 prog_name, opt->env_var);
      }
    }

    if (opt->found_opt)
      *(opt->found_opt) = 0;
    if (opt->found_arg)
      *(opt->found_arg) = 0;

    if (opt->has_arg) {
      OPT_ARG *arg = opt->arg;

      if (arg) {
	switch (opt->arg_type) {
	case MU_OPT_BOOL:
	  arg->arg_bool = opt->bool_default;
	  break;
	case MU_OPT_INT:
	  arg->arg_int = opt->int_default;
	  break;
	case MU_OPT_FLOAT:
	  arg->arg_float = opt->float_default;
	  break;
	case MU_OPT_STRING:
	  arg->arg_string = opt->string_default;
	  break;
	case MU_OPT_FILE:
	  arg->arg_file = opt->file_default;
	  break;
	case MU_OPT_DIRECTORY:
	  arg->arg_directory = opt->dir_default;
	  break;
	case MU_OPT_SUBOPT:
	  if (subopts)
	    mu_die(-1, "%s: cannot have subsuboptions\n", prog_name);
	  else
	    init_opts(prog_name, opt->subopts, 1);
	  break;
	default:
	  mu_die(-1, "%s: invalid option type (type code %u)\n",
		 prog_name, opt->arg_type);
	}
      }

      if (opt->arg_type != MU_OPT_SUBOPT && opt->argstr)
	*(opt->argstr) = NULL;
    }
  }
}

/*
 * Try to parse an option as a short option, and if that doesn't work,
 * try parsing it as a long option instead. Returns 0 on sucess if
 * `argument' wasn't used, 1 on success if `argument' was used, 2 if
 * it's not an option, and an error code if a fatal error occured (in
 * which case an error message will be printed to
 * `stderr'). `argument' and `optstr' must be writable because we
 * might pass them to mu_parse_subopts().
 */
static int parse_opt(const char *prog_name, char *optstr,
		     char *argument, const MU_OPT *options, int flags) {
  int ret;

  /* Try to parse it as a short option. */
  ret = parse_short_opt(prog_name, optstr, argument, options,
			!(flags & MU_OPT_BUNDLE), flags);

  if (ret == 4) {
    /* Long options should take precedence, but we should still parse
       it as a short option if no long option was found. */
    ret = parse_long_opt(prog_name, optstr, argument, options, 1, flags);
    if (ret == 2)
      ret = parse_short_opt(prog_name, optstr, argument,
			    options, 0, flags);
  }
  else if (ret == 2) {
    /* Didn't work? Try to parse it as a long option. */
    ret = parse_long_opt(prog_name, optstr, argument, options, 0, flags);
  }

  if (ret == 3) {
    /* 3 means that MU_OPT_ALLOW_INVALID was passed in flags, so we
       should treat invalid options as positional arguments. */
    assert(flags & MU_OPT_ALLOW_INVALID);
    ret = 2;
  }

  return ret;
}

/*
 * Parse a short option, returning 0 on sucess if `argument' wasn't
 * used, 1 on success if `argument' was used, 2 if it's not a short
 * option, 3 if it's an invalid option and MU_OPT_ALLOW_INVALID was
 * passed in `flags', and an error code if a fatal error occured (in
 * which case an error message will be printed to `stderr').
 * `argument' and `optstr' must be writable because we might pass them
 * to mu_parse_subopts().
 *
 * If `check_req' is nonzero, then parse_short_opt() will return 4 if
 * it matched a short option with a required argument, but it will not
 * call handle_option(). This is used by parse_opt() to determine
 * whether short or long options should take precedence when
 * MU_OPT_BUNDLE is not passed in `flags'.
 */
static int parse_short_opt(const char *prog_name, char *optstr,
			   char *argument, const MU_OPT *options,
			   int check_req, int flags) {
  int used_arg = 0;

  if (*(optstr++) != '-' || *optstr == '-' || !*optstr)
    return 2;

  for (char c = *optstr; (c = *optstr); optstr++) {
    const MU_OPT *opt;
    char *argstr = NULL;
    int ret;

    /* Find the short option. */
    for (opt = options; !OPT_END(opt); opt++) {
      if (opt->short_opt == c)
	break;
    }

    if (OPT_END(opt)) {
      if (!(flags & MU_OPT_BUNDLE))
	return 2;		/* It could still be a long option. */

      if (flags & MU_OPT_ALLOW_INVALID)
	return 3;
      fprintf(stderr, "%s: '-%c': invalid option\n", prog_name, c);
      return MU_OPT_ERR_PARSE;
    }

    /* This is so that parse_short_opt() knows that long options
       should take precedence. But single-character options should
       always be treated as short options. */
    if (check_req && opt->has_arg == MU_OPT_REQUIRED && optstr[1])
      return 4;

    /* Get the argument if there is one. */
    if (optstr[1]) {
      if (opt->has_arg)
	argstr = optstr + 1;
      else if (!(flags & MU_OPT_BUNDLE))
	return 2;
    }
    else if (opt->has_arg == MU_OPT_REQUIRED) {
      argstr = argument;
      used_arg = 1;
    }

    ret = handle_option(prog_name, OPT_TYPE_SHORT, opt, argstr);
    if (MU_OPT_ERR(ret))
      return ret;

    if (!(flags & MU_OPT_BUNDLE) || opt->has_arg)
      break;
  }

  return used_arg;
}

/*
 * Parse a long option, returning 0 on sucess if `argument' wasn't
 * used, 1 on success if `argument' was used, 2 if it's not a long
 * option, 3 if it's an invalid option and MU_OPT_ALLOW_INVALID was
 * passed in `flags', and an error code if a fatal error occured (in
 * which case an error message will be printed to
 * `stderr'). `argument' must be writable because we might pass it to
 * mu_parse_subopts(). `could_be_short' should be nonzero if the
 * option might actually be a short option.
 */
static int parse_long_opt(const char *prog_name, const char *optstr,
			  char *argument, const MU_OPT *options,
			  int could_be_short, int flags) {
  int used_arg = 0, ret;
  unsigned *matches, n_matches;
  const MU_OPT *opt;
  char *argstr;	/* Must be writable because we might pass it to
		   mu_parse_subopts(). */
  const char *dash;
  int exact_match;

  if (*(optstr++) != '-')
    return 2;

  /* If we're not bundling options, long options can start with a
     single '-' too (or a "--"). */
  if (*optstr == '-') {
    optstr++;
    dash = "--";
    could_be_short = 0;
  }
  else if (flags & MU_OPT_BUNDLE)
    return 2;
  else
    dash = "-";

  if (!*optstr)
    return 2;

  /* An option cannot start with ARG_SEPARATOR. */
  if (*optstr == ARG_SEPARATOR) {
    /* Don't check if MU_OPT_ALLOW_INVALID is in `flags', because
       ARG_SEPARATOR should *never* appear in a long option. */
    fprintf(stderr, "%s: '%s%s': invalid option\n",
	    prog_name, dash, optstr);
    return MU_OPT_ERR_PARSE;
  }

  /* Search for the option. */
  n_matches =
    find_unambiguously(&matches, &exact_match, options, optstr, 0);
  if (could_be_short && !exact_match) {
    /* It didn't match exactly, so treat it as a short option. */
    free(matches);
    return 2;
  }
  if (n_matches != 1) {
    if (flags & MU_OPT_ALLOW_INVALID) {
      free(matches);
      return 3;
    }

    if (n_matches) {
      fprintf(stderr, "%s: '%s%s': option is ambiguous; possibilities:\n",
	      prog_name, dash, optstr);
      /* Sort the matches so they appear in the same order as they do
	 in the help message. */
      qsort(matches, n_matches, sizeof(*matches), index_sort);
      for (unsigned i = 0; i < n_matches; i++)
	fprintf(stderr, "  %s%s\n", dash, options[matches[i]].long_opt);
    }
    else
      fprintf(stderr, "%s: '%s%s': invalid option\n",
	      prog_name, dash, optstr);

    free(matches);		/* Come get them while they last! */

    return MU_OPT_ERR_PARSE;
  }

  opt = options + *matches;

  free(matches);		/* Seriously! */

  argstr = strchr(optstr, ARG_SEPARATOR);
  if (argstr)
    argstr++;
  else if (opt->has_arg == MU_OPT_REQUIRED) {
    argstr = argument;
    used_arg = 1;
  }

  ret = handle_option(prog_name, OPT_TYPE_LONG, opt, argstr, dash);
  return MU_OPT_ERR(ret) ? ret : used_arg;
}

/*
 * Parse an environment variable. Note that, unlike options, if an
 * unknown environment variable is encountered, it is not treated as
 * an error. Also note that, unlike long options and suboptions,
 * abbreviation is not allowed.
 */
static int parse_env_var(const char *prog_name, const char *env_var,
			 const MU_OPT *options, int flags) {
  const char *equals;
  size_t eq_ind;
  int ret = 0;

  equals = strchr(env_var, ARG_SEPARATOR);
  if (!equals || equals == env_var)
    mu_die(-1, "invalid environment variable: '%s'\n", env_var);
  eq_ind = equals - env_var;

  for (const MU_OPT *opt = options; !(ret || OPT_END(opt)); opt++) {
    ret = check_env_var(prog_name, opt, env_var, eq_ind);

    if (!ret && opt->has_arg && opt->arg_type == MU_OPT_SUBOPT) {
      /* Search the suboptions as well. */
      for (const MU_OPT *subopt = opt->subopts;
	   !(ret || SUBOPT_END(subopt)); subopt++)
	ret = check_env_var(prog_name, subopt, env_var, eq_ind);
    }
  }

  return MU_OPT_ERR(ret) ? ret : 0;
}

/*
 * Check if `opt' matches the environment variable `env', and call
 * handle_option() if it does. `eq_ind' is the index of the argument
 * separator ('=') in `env_var', so we don't have to search for it
 * every time we're called. Returs zero if it doesn't match, one if it
 * does match, or an error code if an error occured.
 */
static int check_env_var(const char *prog_name, const MU_OPT *opt,
			 const char *env_var, size_t eq_ind) {
  if (opt->env_var && strlen(opt->env_var) == eq_ind &&
      !strncmp(opt->env_var, env_var, eq_ind)) {
    const char *arg;
    char *arg_copy;
    int ret;

    arg = env_var + eq_ind + 1;
    if (*arg) {
      /* Make a copy because if there are suboptions, we don't want to
	 actually modify the environment. However, we don't want to
	 make a copy if the argument is not suboptions, because then
	 when we free the copy, '*(opt->argstr)' will be rendered
	 invalid. */
      if (opt->has_arg && opt->arg_type == MU_OPT_SUBOPT)
	arg_copy = mu_xstrdup(arg);
      else {
	/* Since we know that 'opt->arg_type' is not MU_OPT_SUBOPT
	   (and thus handle_option() won't modify `arg_copy'), we can
	   perform this cast safely. */
	arg_copy = (char *)arg;
      }
    }
    else
      arg_copy = NULL;

    ret = handle_option(prog_name, OPT_TYPE_ENV, opt, arg_copy);

    if (arg_copy && arg_copy != arg)
      free(arg_copy);

    return MU_OPT_ERR(ret) ? ret : 1;
  }

  return 0;
}

/*
 * Handle an option. Returns zero on success, or an error code on
 * error (in which case an error message will be printed to `stderr').
 */
static int handle_option(const char *prog_name, enum opt_type opt_type,
			 const MU_OPT *opt, char *argstr, ...) {
  OPT_ARG arg = { 0 };
  OPT_ARG *p_arg = opt->arg ? opt->arg : &arg;
  int cb_ret;
  const char *dash;

  if (opt_type == OPT_TYPE_LONG) {
    va_list ap;
    va_start(ap, argstr);
    dash = va_arg(ap, const char *);
    va_end(ap);
  }

  if (opt->has_arg == MU_OPT_REQUIRED && !argstr) {
    switch (opt_type) {
    case OPT_TYPE_SHORT:
      fprintf(stderr, "%s: '-%c': option requires argument\n",
	      prog_name, opt->short_opt);
      break;
    case OPT_TYPE_LONG:
      fprintf(stderr, "%s: '%s%s': option requires argument\n",
	      prog_name, dash, opt->long_opt);
      break;
    case OPT_TYPE_ENV:
      fprintf(stderr,
	      "%s: environment variable '%s' requires value\n",
	      prog_name, opt->env_var);
      break;
    case OPT_TYPE_SUBOPT:
      fprintf(stderr, "%s: '%s': suboption requires argument\n",
	      prog_name, opt->subopt_name);
      break;
    }

    return MU_OPT_ERR_PARSE;
  }

  /*
   * Process the argument.
   */
  if (argstr) {
    if (opt->has_arg == MU_OPT_NONE) {
      switch (opt_type) {
      case OPT_TYPE_SHORT:
	fprintf(stderr, "%s: '-%c': option does not take argument\n",
		prog_name, opt->short_opt);
	break;
      case OPT_TYPE_LONG:
	fprintf(stderr, "%s: '%s%s': option does not take argument\n",
		prog_name, dash, opt->long_opt);
	break;
      case OPT_TYPE_ENV:
	fprintf(stderr,
		"%s: environment variable '%s' cannot have value\n",
		prog_name, opt->env_var);
	break;
      case OPT_TYPE_SUBOPT:
	fprintf(stderr, "%s: '%s': suboption does not take argument\n",
		prog_name, opt->subopt_name);
	break;
      }

      return MU_OPT_ERR_PARSE;
    }

    /* We are going to overwrite the previous argument. */
    if (free_opt_arg(prog_name, opt))
      return 1;

    switch (opt->arg_type) {
      size_t arglen;
      long arg_int;
      char *endptr;
      int ret;

    case MU_OPT_BOOL:
      arglen = strlen(argstr);
      if (!strncasecmp(argstr, "yes", arglen) ||
	  !strncasecmp(argstr, "true", arglen)) {
	p_arg->arg_bool = 1;
	break;
      }
      if (!strncasecmp(argstr, "no", arglen) ||
	  !strncasecmp(argstr, "false", arglen)) {
	p_arg->arg_bool = 0;
	break;
      }

      /* fall-through */

    case MU_OPT_INT:
      errno = 0;
      arg_int = strtol(argstr, &endptr, 0);

      if (*endptr) {
	fprintf(stderr, "%s: %s: not a%s\n", prog_name, argstr,
		(opt->arg_type == MU_OPT_BOOL) ?
		" yes or no value" : "n integer");
	return MU_OPT_ERR_PARSE;
      }

      if (opt->arg_type == MU_OPT_INT) {
	if (!errno && (arg_int < opt->ibound.lower ||
		       arg_int > opt->ibound.upper))
	  errno = ERANGE;

	if (errno) {
	  fprintf(stderr, "%s: %s: %s\n", prog_name, argstr, strerror(errno));
	  return MU_OPT_ERR_PARSE;
	}
      }

      if (opt->arg_type == MU_OPT_BOOL)
	p_arg->arg_bool = !!arg_int;
      else
	p_arg->arg_int = arg_int;

      break;

    case MU_OPT_FLOAT:
      errno = 0;
      p_arg->arg_float = strtod(argstr, &endptr);

      if (*endptr) {
	fprintf(stderr, "%s: %s: not a floating-point number\n",
		prog_name, argstr);
	return MU_OPT_ERR_PARSE;
      }

      if (!errno && (p_arg->arg_float < opt->fbound.lower ||
		     p_arg->arg_float > opt->fbound.upper))
	errno = ERANGE;

      if (errno) {
	fprintf(stderr, "%s: %s: %s\n", prog_name, argstr, strerror(errno));
	return MU_OPT_ERR_PARSE;
      }

      break;

    case MU_OPT_STRING:
      p_arg->arg_string = argstr;
      break;

    case MU_OPT_FILE:
      if (strcmp(argstr, "-"))
	p_arg->arg_file = fopen(argstr, opt->file_mode);
      else {
	char c;
	const char *mode = opt->file_mode;

	/* Handle stdin/stdout. */
	c = *(mode++);
	if (*mode == 'b')
	  mode++;
	if (*mode) {
	  /* Even if it's '+', it's still not valid for
	     stdin/stdout. */
	  errno = EINVAL;
	  p_arg->arg_file = NULL;
	}

	switch (c) {
	case 'r':
	  p_arg->arg_file = stdin;
	  argstr = "<stdin>";
	  break;

	case 'w':
	case 'a':
	  p_arg->arg_file = stdout;
	  argstr = "<stdout>";
	  break;

	default:
	  errno = EINVAL;
	  p_arg->arg_file = NULL;
	}
      }

      if (!p_arg->arg_file) {
	fprintf(stderr, "%s: cannot open %s: %s\n",
		prog_name, argstr, strerror(errno));
	return MU_OPT_ERR_IO;
      }

      break;

    case MU_OPT_DIRECTORY:
      p_arg->arg_directory = opendir(argstr);
      if (!p_arg->arg_directory) {
	fprintf(stderr, "%s: cannot open directory %s: %s\n",
		prog_name, argstr, strerror(errno));
	return MU_OPT_ERR_IO;
      }

      break;

    case MU_OPT_SUBOPT:
      if (opt_type == OPT_TYPE_SUBOPT)
	mu_die(-1, "%s: cannot have subsuboptions\n", prog_name);

      ret = mu_parse_subopts(prog_name, argstr, opt->subopts);
      if (ret)
	return ret;

      break;

    default:
      mu_die(-1, "%s: invalid option type (type code %u)\n",
	     prog_name, opt->arg_type);
    }
  }

  if (opt->found_opt)
    *(opt->found_opt) = 1;
  if (opt->found_arg)
    *(opt->found_arg) = !!argstr;

  if (opt->arg_type != MU_OPT_SUBOPT && opt->argstr)
    *(opt->argstr) = argstr;

  /*
   * Call the callback.
   */
  if (opt->callback_none) {
    /* Initialize to zeros in case the callback forgets to add a
       terminating null byte. +1 for the terminating null byte. */
    char err[MU_OPT_ERR_MAX + 1] = { 0 };

    if (opt->has_arg == MU_OPT_NONE) {
      cb_ret = opt->callback_none(opt->cb_data, err);
    }
    else {
      switch (opt->arg_type) {
      case MU_OPT_BOOL:
	cb_ret =
	  opt->callback_bool(!!argstr, p_arg->arg_bool,
			     opt->cb_data, err);
	break;
      case MU_OPT_INT:
	cb_ret =
	  opt->callback_int(!!argstr, p_arg->arg_int,
			    opt->cb_data, err);
	break;
      case MU_OPT_FLOAT:
	cb_ret =
	  opt->callback_float(!!argstr, p_arg->arg_float,
			      opt->cb_data, err);
	break;
      case MU_OPT_STRING:
	cb_ret =
	  opt->callback_string(!!argstr, p_arg->arg_string,
			       opt->cb_data, err);
	break;
      case MU_OPT_FILE:
	cb_ret =
	  opt->callback_file(!!argstr, argstr, p_arg->arg_file,
			     opt->cb_data, err);
	break;
      case MU_OPT_DIRECTORY:
	cb_ret =
	  opt->callback_directory(!!argstr, argstr, p_arg->arg_directory,
				  opt->cb_data, err);
	break;
      case MU_OPT_SUBOPT:
	cb_ret = 0;	    /* Suboptions have there own callbacks. */
	break;
      default:
	mu_die(-1, "%s: invalid option type (type code %u)\n",
	       prog_name, opt->arg_type);
      }
    }

    if (cb_ret)
      fprintf(stderr, "%s: %s\n", prog_name, err);
  }
  else
    cb_ret = 0;

  if (argstr && !opt->arg) {
    /* So we can modify the `arg' and `argstr' fields. */
    MU_OPT mutable_opt = *opt;
    const char *const_argstr = argstr;
    int ret;

    mutable_opt.arg = p_arg;
    mutable_opt.argstr = &const_argstr;

    /* We're not going to return the file/directory we opened, so we
       should close it now. */
    ret = free_opt_arg(prog_name, &mutable_opt);
    if (ret)
      return ret;
  }

  return cb_ret ? MU_OPT_ERR_CALLBACK : 0;
}

/*
 * Find the long option `needle' in `haystack', allowing abbreviations
 * as long as they are not ambiguous. `subopts' should be nonzero if
 * the options in `haystack' are suboptions, or 0 if they
 * aren't. Returns the number of options that were found in
 * `haystack'. `matches' will be allocated and will be a list of
 * indices in `haystack' that point to the found options. If
 * `exact_match' is not NULL, '*exact_match' will be set to nonzero if
 * `needle' matched exactly (not just unambiguously), or 0 if it
 * didn't.
 */
static unsigned find_unambiguously(unsigned **matches, int *exact_match,
				   const MU_OPT *haystack,
				   const char *needle, int subopts) {
  unsigned n_haystack, n_matches, index;
  size_t needle_size;
  const MU_OPT *opt;

  /* Get the size of the matches and the haystack. */
  n_matches = 0;
  for (n_haystack = 0, opt = haystack;
       !ANYOPT_END(opt, subopts); n_haystack++, opt++) {
    if (subopts ? opt->subopt_name : opt->long_opt)
      n_matches++;
  }

  /* Construct the matches. */
  *matches = mu_xmalloc(sizeof(**matches) * n_matches);
  index = 0;
  for (unsigned i = 0; i < n_haystack; i++) {
    opt = haystack + i;
    if (subopts ? opt->subopt_name : opt->long_opt) {
      assert(index < n_matches);
      (*matches)[index++] = i;
    }
  }
  assert(index == n_matches);

  /* We stop parsing at ARG_SEPARATOR. */
  needle_size = strchrnul(needle, ARG_SEPARATOR) - needle;

  /* Search for the needle. */
  if (exact_match)
    *exact_match = 0;
  for (size_t i = 0; i < needle_size; i++) {
    /* Remove all non-matching items. */
    for (unsigned j = 0; j < n_matches;) {
      const char *check;

      opt = haystack + (*matches)[j];
      check = subopts ? opt->subopt_name : opt->long_opt;
      if (check[i] != needle[i])
	(*matches)[j] = (*matches)[--n_matches];
      else if (i + 1 == needle_size && !check[i + 1]) {
	/* There is only one possibility, so it's not ambiguous
	   (e.g. "foo" matches "foo" unambiguously out of "foo" and
	   "foo-bar"). */
	**matches = (*matches)[j];
	if (exact_match)
	  *exact_match = 1;
	return 1;
      }
      else
	j++;
    }

    /* There's no point in searching any more if all matches are
       gone. */
    if (!n_matches)
      break;
  }

  return n_matches;
}

/*
 * Compare two `unsigned's. Pass this to qsort().
 */
static int index_sort(const void *ap, const void *bp) {
  unsigned a = *(unsigned *)ap, b = *(unsigned *)bp;
  if (a < b)
    return -1;
  else if (a == b)
    return 0;
  else
    return 1;
}

/*
 * Return the argument text for an option or return a string based on
 * the type of the argument if there is no argument text.
 */
static const char *get_arg_help(const MU_OPT *opt) {
  if (opt->arg_help)
    return opt->arg_help;

  switch (opt->arg_type) {
  case MU_OPT_BOOL:
    return "Y/N";
  case MU_OPT_INT:
    return "INT";
  case MU_OPT_FLOAT:
    return "FLOAT";
  case MU_OPT_STRING:
    return "STRING";
  case MU_OPT_FILE:
    return "FILE";
  case MU_OPT_DIRECTORY:
    return "DIR";
  case MU_OPT_SUBOPT:
    return "SUBOPTS";
  default:
    mu_die(-1, "Invalid option type (type code %u)\n", opt->arg_type);
  }
}

/*
 * Return the string that should be used to separate an option and its
 * argument. `type' is the type of the option.
 */
static const char *get_arg_separator(const MU_OPT *opt,
				     enum opt_type type) {
  switch (type) {
  case OPT_TYPE_SHORT:
  case OPT_TYPE_LONG:
    if (opt->long_opt == NULL)
      break;
    /* fall-through */
  case OPT_TYPE_ENV:
  case OPT_TYPE_SUBOPT:
    return "=";

  default:
    mu_die(-1, "Invalid option type (code %u)\n", type);
  }

  return (opt->has_arg == MU_OPT_OPTIONAL) ? "" : " ";
}

/*
 * Returns the minimum amount the help text should be indented for
 * this option. Set `short_opts' to nonzero if some of the options are
 * short.
 */
static size_t get_min_indent(const MU_OPT *opt, enum opt_type type,
			     int has_short_opts) {
  size_t opt_len;

  switch (type) {
  case OPT_TYPE_SHORT:
  case OPT_TYPE_LONG:
    if (opt->short_opt)
      opt_len = 2;		/* +1 for '-' */

    if (opt->long_opt) {
      opt_len = strlen(opt->long_opt) + 2; /* +2 for "--" */
      if (has_short_opts)
	opt_len += 4;		/* For "-s, " or "    " */
    }

    break;

  case OPT_TYPE_ENV:
    opt_len = strlen(opt->env_var);
    break;

  case OPT_TYPE_SUBOPT:
    opt_len = strlen(opt->subopt_name);
    break;
  }

  if (opt->has_arg) {
    /* Add the length of the argument string. +1 for ' ' or '='. */
    opt_len += strlen(get_arg_help(opt)) + 1;
    if (opt->has_arg == MU_OPT_OPTIONAL)
      opt_len += 2;		/* For '[' and ']' */
  }

  /* For the two leading spaces and an extra two spaces between the
     option and its help. */
  opt_len += 4;

  return opt_len;
}

/*
 * Release resources (currently just opened files/directories)
 * associated with 'opt->arg'. Returns zero on success or an error
 * code on error, in which case an error message will be printed to
 * `stderr'.
 */
static int free_opt_arg(const char *prog_name, const MU_OPT *opt) {
  OPT_ARG *arg = opt->arg;

  if (!arg)
    return 0;

  switch (opt->arg_type) {
  case MU_OPT_BOOL:
    arg->arg_bool = 0;
    break;
  case MU_OPT_INT:
    arg->arg_int = 0;
    break;
  case MU_OPT_FLOAT:
    arg->arg_float = 0;
    break;
  case MU_OPT_STRING:
    arg->arg_string = NULL;
    break;

  case MU_OPT_FILE:
    if (arg->arg_file) {
      if (arg->arg_file != stdin && arg->arg_file != stdout &&
	  fclose(arg->arg_file)) {
	fprintf(stderr, "%s: cannot close %s: %s\n", prog_name,
		opt->argstr ? *(opt->argstr) : "file", strerror(errno));
	return MU_OPT_ERR_IO;
      }
      arg->arg_file = NULL;
    }
    break;

  case MU_OPT_DIRECTORY:
    if (arg->arg_directory) {
      if (closedir(arg->arg_directory)) {
	fprintf(stderr, "%s: cannot close directory%s%s: %s\n", prog_name,
		opt->argstr ? " " : "", opt->argstr ? *(opt->argstr) : "",
		strerror(errno));
	return MU_OPT_ERR_IO;
      }
      arg->arg_directory = NULL;
    }

    break;

  case MU_OPT_SUBOPT:
    break;

  default:
    mu_die(-1, "%s: invalid option type (type code %u)\n",
	   prog_name, opt->arg_type);
  }

  if (opt->arg_type != MU_OPT_SUBOPT && opt->argstr)
    *(opt->argstr) = NULL;

  return 0;
}

/*
 * Call free_opt_arg() for each option in `options'. Returns the
 * number of failures.
 */
static int free_opt_args(const char *prog_name, const MU_OPT *options) {
  int ret = 0;

  for (const MU_OPT *opt = options; !OPT_END(opt); opt++) {
    if (opt->has_arg && opt->arg_type == MU_OPT_SUBOPT) {
      for (const MU_OPT *subopt = opt->subopts;
	   !SUBOPT_END(subopt); subopt++)
	ret += free_opt_arg(prog_name, subopt);
    }
    else
      ret += free_opt_arg(prog_name, opt);
  }

  return ret;
}

/*
 * Format help output from `usage' and `options', using `format' to
 * format the message. `data' will be passed to `format' as the first
 * argument. Returns zero on success or nonzero on error, in which
 * case `errno' will be set to indicate the error.
 */
static int format_help_plain(const char *prog_name, char **usage,
			     const char *desc, const char *notes,
			     const MU_OPT *options,
			     FORMAT_HELP_CALLBACK format, void *data,
			     unsigned short goal,
			     unsigned short width) {
  unsigned short indent, offset, max_indent, cur_col = 0;
  int short_opts, env_vars, print_mandatory_args_msg;
  const MU_OPT *opt, *subopt;
  int first;

  /* The amount we should indent beyond the indentation of the first
     line for following lines. */
  offset = 2;

  /* Give the help text at least 5/8 of the screen (so the options get
     3/8). */
  max_indent = (goal * 3) / 8;
  /* This is possible for very small goals/widths. */
  if (max_indent + offset >= width - 1) {
    if (max_indent >= width)
      max_indent = width - 1;
    if (width - max_indent - 2 >= 0)
      offset = width - max_indent - 2; /* Allow for a continuing '-' */
    else {
      offset = width - max_indent - 1;
      /* Allow for a continuing '-' if there's room. */
      if (max_indent > 0)
	max_indent--;
    }
  }

  assert(prog_name);

  /* Format the usage(s). */
  assert(*usage);
  for (first = 1; *usage; usage++, first = 0) {
    if (format(data, 0, 0, &cur_col, 0, 0, "%s %s%s%s\n",
	       first ? "Usage:" : "  or: ",
	       prog_name, **usage ? " " : "", *usage)) {
      return 1;
    }
  }

  if ((desc && format(data, goal, width, &cur_col, 0, 0, "%s\n", desc)) ||
      format(data, goal, width, &cur_col, 0, 0, "%s",
	     OPT_END(options) ? "" : "\n"))
    return 1;

 /*
  * Check whether there are long options with requried arguments that
  * have equivalent short options, and also whether there are short
  * options and environment variables at all.
  */
  short_opts = env_vars = print_mandatory_args_msg = 0;
  for (opt = options; !OPT_END(opt); opt++) {
    if (opt->short_opt) {
      short_opts = 1;
      if (opt->long_opt && opt->has_arg == MU_OPT_REQUIRED)
	print_mandatory_args_msg = 1;
    }

    if (opt->env_var)
      env_vars = 1;
    if (!env_vars && opt->has_arg && opt->arg_type == MU_OPT_SUBOPT) {
      for (subopt = opt->subopts; !SUBOPT_END(subopt); subopt++) {
	if (subopt->env_var) {
	  env_vars = 1;
	  break;
	}
      }
    }

    if (short_opts && env_vars && print_mandatory_args_msg)
      /* We have all the info we need. */
      break;
  }

  /* Get the amount we should indent. */
  indent = 0;
  for (opt = options; !OPT_END(opt); opt++) {
    unsigned short opt_len, max_opt_len = 0;

    if (opt->short_opt || opt->long_opt) {
      opt_len = get_min_indent(opt, OPT_TYPE_LONG, short_opts);
      if (opt_len > max_opt_len)
	max_opt_len = opt_len;
    }
    if (opt->env_var) {
      opt_len = get_min_indent(opt, OPT_TYPE_ENV, 0);
      if (opt_len > max_opt_len)
	max_opt_len = opt_len;
    }

    /* Don't count it if it's more than the maximum indent. */
    if (max_opt_len > indent && max_opt_len <= max_indent)
      indent = max_opt_len;

    if (opt->has_arg && opt->arg_type == MU_OPT_SUBOPT) {
      /* Don't forget the suboptions. */
      for (subopt = opt->subopts; !SUBOPT_END(subopt); subopt++) {
	opt_len = get_min_indent(subopt, OPT_TYPE_SUBOPT, 0);
	if (opt_len > max_opt_len)
	  max_opt_len = opt_len;
	if (subopt->env_var) {
	  opt_len = get_min_indent(subopt, OPT_TYPE_ENV, 0);
	  if (opt_len > max_opt_len)
	    max_opt_len = opt_len;
	}

	/* Don't count it if it's more than the maximum indent. */
	if (max_opt_len > indent && max_opt_len <= max_indent)
	  indent = max_opt_len;
      }
    }
  }

  /* The indent should be at least 4. */
  if (indent < 4) {
    indent = 4;
    /* Unless, of course, it's greater than the maximum indent. */
    if (indent > max_indent)
      indent = max_indent;
  }

  if (print_mandatory_args_msg &&
      format(data, goal, width, &cur_col, 0, 0,
	     MANDATORY_ARGS_MSG "\n"))
    return 1;

  /* Print the help for each option. */
  for (opt = options; !OPT_END(opt); opt++) {
    if (format_opt_help(opt, OPT_TYPE_LONG, short_opts,
			format, data, goal, width,
			&cur_col, indent, offset)) {
      return 1;
    }
  }

  /* Print suboption help after regular options. */
  for (opt = options; !OPT_END(opt); opt++) {
    if (opt->has_arg && opt->arg_type == MU_OPT_SUBOPT) {
      if (print_subopts_header(opt, "\n", ":\n", "-",
			       format, data, &cur_col, goal, width))
	return 1;

      for (subopt = opt->subopts; !SUBOPT_END(subopt); subopt++) {
	if (format_opt_help(subopt, OPT_TYPE_SUBOPT, 0,
			    format, data, goal, width,
			    &cur_col, indent, offset)) {
	  return 1;
	}
      }
    }
  }

  /* Finally, print environment variable help. */
  if (env_vars) {
    if (format(data, goal, width, &cur_col, 0, 0,
	       "\nENVIRONMENT\n\n"))
      return 1;

    for (opt = options; !OPT_END(opt); opt++) {
      if (format_opt_help(opt, OPT_TYPE_ENV, 0,
			  format, data, goal, width,
			  &cur_col, indent, offset)) {
	return 1;
      }

      if (opt->has_arg && opt->arg_type == MU_OPT_SUBOPT) {
	/* Print environment variables for suboptions as well. */
	for (subopt = opt->subopts; !SUBOPT_END(subopt); subopt++) {
	  if (format_opt_help(subopt, OPT_TYPE_ENV, 0,
			      format, data, goal, width,
			      &cur_col, indent, offset)) {
	    return 1;
	  }
	}
      }
    }
  }

  if (notes &&
      format(data, goal, width, &cur_col, 0, 0, "\n%s\n", notes))
    return 1;

  return 0;
}

/*
 * Like format_help_plain(), but output help in the man(7)
 * format. Returns zero on success or nonzero on error, in which case
 * `errno' will be set to indicate the error.
 */
static int format_help_man(const char *prog_name, char **usage,
			   const char *short_desc, const char *desc,
			   const char *notes, const char *section,
			   const char *sect_name, const char *pkg_str,
			   const char *date, const MU_OPT *options,
			   FORMAT_HELP_CALLBACK format, void *data) {
  unsigned short cur_col = 0;
  int env_vars, print_mandatory_args_msg;
  char *tmp, *title_name, *name;
  const MU_OPT *opt, *subopt;
  int ret;
  /* Used when `date' is NULL. 19 bytes for the year, 2 for the month,
     2 for the date, 2 for the hyphens, and 1 for the terminating null
     byte. The reason we use 19 bytes for the year is because, if
     sizeof(int) is 8, the maximum year is 2^(8*8 - 1) + 1900 which
     requires 19 bytes to be represented in decimal. This way, we
     won't fail even if the user has their clock set to some crazy
     date (or, unlikely, if it actually is that date). */
  char date_str[26];
  const char *const sect_table[] = {
    NULL,
    /* Default section names. */
    "User Commands",		/* 1 */
    "System Calls",		/* 2 */
    "Library Calls",		/* 3 */
    "Special Files",		/* 4 */
    "File Formats",		/* 5 */
    "Games Manual",		/* 6 */
    /* No default for section 7 because that can be lots of things. */
    NULL,			/* 7 */
    "System Administration",	/* 8 */
    /* This one's non standard, but there's no reason not to support
       it. */
    "Kernel Routines"		/* 9 */
  };

  if (!sect_name) {
    int sect_index;

    sect_index = atoi(section);
    if (sect_index < 0 ||
	sect_index >= sizeof(sect_table) / sizeof(*sect_table)) {
      errno = ERANGE;
      return 1;
    }

    sect_name = sect_table[sect_index];
    if (!sect_name) {
      errno = EINVAL;
      return 1;
    }
  }

  if (!date) {
    struct timespec secs;
    struct tm *calendar_time;

    /* Get the current time. */
    if (clock_gettime(CLOCK_REALTIME, &secs))
      return 1;

    /* Convert it to calendar time. */
    calendar_time = localtime(&(secs.tv_sec));
    if (!calendar_time)
      return 1;

    /* And convert that to a string. */
    if (!strftime(date_str, sizeof(date_str),
		  "%Y-%m-%d", calendar_time)) {
      /* This should not be possible because `date_str' should big
	 enough to hold all possible dates. */
      mu_die(-1, "date overflowed (this should not be happening)");
    }

    date = date_str;
  }

  /*
   * Check whether there are long options with requried arguments that
   * have equivalent short options, and also whether there are
   * environment variables.
   */
  env_vars = print_mandatory_args_msg = 0;
  for (opt = options; !OPT_END(opt); opt++) {
    if (opt->short_opt && opt->long_opt &&
	opt->has_arg == MU_OPT_REQUIRED)
      print_mandatory_args_msg = 1;

    if (opt->env_var)
      env_vars = 1;
    if (!env_vars && opt->has_arg && opt->arg_type == MU_OPT_SUBOPT) {
      for (subopt = opt->subopts; !SUBOPT_END(subopt); subopt++) {
	if (subopt->env_var) {
	  env_vars = 1;
	  break;
	}
      }
    }

    if (env_vars && print_mandatory_args_msg)
      /* We have all the info we need. */
      break;
  }

  if (format(data, 0, 0, &cur_col, 0, 0,
	     ".\\\" DO NOT MODIFY THIS FILE! Generated by %s.\n",
	     PACKAGE_STRING))
    return 1;

  /* Get the basename of `prog_name'. Make a copy because basename(3)
     is allowed to modify its argument. */
  tmp = strdup(prog_name);
  if (!tmp)
    return 1;
  name = strdup(basename(tmp));
  free(tmp);
  if (!name)
    return 1;

  /* Get the name to use in the title (same as `name' but ALL
     CAPS). */
  title_name = strdup(name);
  if (!title_name) {
    free(name);
    return 1;
  }
  for (char *p = title_name; *p; p++)
    *p = toupper(*p);

  if (!pkg_str)
    pkg_str = name;

  /* Print the title. */
  ret = format(data, 0, 0, &cur_col, 0, 0,
	       ".TH %s \"%s\" \"%s\" \"%s\" \"%s\"\n",
	       title_name, section, date, pkg_str, sect_name);
  free(title_name);
  if (ret) {
    free(name);
    return 1;
  }

  if (short_desc)
    tmp = strdup(short_desc);
  else if (asprintf(&tmp, "manual page for %s", pkg_str) < 0)
    tmp = NULL;
  if (!tmp) {
    free(name);
    return 1;
  }

  /* Print the name and short description. */
  ret = format(data, 0, 0, &cur_col, 0, 0,
	       ".SH NAME\n%s \\- %s\n", name, tmp);
  free(tmp);
  if (ret) {
    free(name);
    return 1;
  }

  /* Print the synopsis section header. */
  if (format(data, 0, 0, &cur_col, 0, 0, ".SH SYNOPSIS\n")) {
    free(name);
    return 1;
  }

  /* Format the possible usages. */
  assert(*usage);
  for (; *usage; usage++) {
    tmp = format_man_str(*usage, 0);
    ret = !tmp ||
      format(data, MAN_PARAGRAPH_WIDTH, 0, &cur_col, 0, 0,
	     ".B %s%s%s\n%s", name, *tmp ? "\n" : "", tmp,
	     usage[1] ? ".br\n" : "");
    free(tmp);
    if (ret)
      break;
  }
  free(name);
  if (ret)
    return 1;

  /* Print the description section header. */
  if (format(data, 0, 0, &cur_col, 0, 0, ".SH DESCRIPTION\n"))
    return 1;

  if (desc) {
    tmp = format_man_str(desc, 0);
    if (!tmp)
      return 1;

    /* Print the description. */
    ret = format(data, MAN_PARAGRAPH_WIDTH, 0, &cur_col, 0, 0,
		 "%s\n", tmp);
    free(tmp);
    if (ret)
      return 1;
  }

  /* Print the mandatory arguments message if we need to.  */
  if (print_mandatory_args_msg &&
      (format(data, 0, 0, &cur_col, 0, 0, ".PP\n") ||
       format(data, MAN_PARAGRAPH_WIDTH, 0, &cur_col, 0, 0,
	      MANDATORY_ARGS_MSG "\n"))) {
    return 1;
  }

  /* Print the help for each option. */
  for (opt = options; !OPT_END(opt); opt++) {
    if (format_opt_help_man(opt, OPT_TYPE_LONG,
			    format, data, &cur_col)) {
      return 1;
    }
  }

  /* Print suboption help after regular options. */
  for (opt = options; !OPT_END(opt); opt++) {
    if (opt->has_arg && opt->arg_type == MU_OPT_SUBOPT) {
      if (print_subopts_header(opt, ".SS \"", "\"\n", "\\-",
			       format, data, &cur_col, 0, 0))
	return 1;

      for (subopt = opt->subopts; !SUBOPT_END(subopt); subopt++) {
	if (format_opt_help_man(subopt, OPT_TYPE_SUBOPT,
				format, data, &cur_col))
	  return 1;
      }
    }
  }

  /* Finally, print environment variable help. */
  if (env_vars) {
    if (format(data, 0, 0, &cur_col, 0, 0, ".SH ENVIRONMENT\n"))
      return 1;

    for (opt = options; !OPT_END(opt); opt++) {
      if (format_opt_help_man(opt, OPT_TYPE_ENV,
			      format, data, &cur_col))
	return 1;

      if (opt->has_arg && opt->arg_type == MU_OPT_SUBOPT) {
	/* Print environment variables for suboptions as well. */
	for (subopt = opt->subopts; !SUBOPT_END(subopt); subopt++) {
	  if (format_opt_help_man(subopt, OPT_TYPE_ENV,
				  format, data, &cur_col))
	    return 1;
	}
      }
    }
  }

  if (notes) {
    tmp = format_man_str(notes, 0);
    if (!tmp)
      return 1;

    /* Print the notes. */
    ret =
      format(data, 0, 0, &cur_col, 0, 0, ".SH NOTES\n") ||
      format(data, MAN_PARAGRAPH_WIDTH, 0, &cur_col, 0, 0,
	     "%s\n", notes);
    free(tmp);
    if (ret)
      return 1;
  }

  return 0;
}

/*
 * Convert a newline separated list of strings to an allocated
 * NULL-terminated list of strings. Returns the allocated list of
 * strings on success or NULL on error, in which case `errno' will be
 * set to indicate the error. The strings in the returned list are
 * dynamically allocated, as well as the list itself.
 */
static char **create_usage_list(const char *str) {
  struct string_list *first, *last;
  size_t size;
  char **out, **pos;

  if (!str) {
    /* Default to "[OPTION]...". */
    out = malloc(sizeof(*out) * 2);
    if (!out)
      return NULL;
    out[0] = strdup("[OPTION]...");
    if (!out[0]) {
      free(out);
      return NULL;
    }
    out[1] = NULL;

    return out;
  }

  first = last = malloc(sizeof(*first));
  if (!first)
    return NULL;

  /* Create a list of strings. */
  size = 0;
  while (1) {
    const char *newline;
    size_t str_size;

    newline = strchrnul(str, '\n');
    str_size = newline - str;

    /* +1 for the terminating null byte. */
    last->str = malloc(str_size + 1);
    if (!last->str) {
      free_string_list(first, last);
      return NULL;
    }
    memcpy(last->str, str, str_size);
    last->str[str_size] = '\0';

    last->next = malloc(sizeof(*(last->next)));
    if (!last->next) {
      free(last->str);
      free_string_list(first, last);
      return NULL;
    }
    last = last->next;
    size++;
    str = newline;
    if (*str)
      str++;
    else
      break;
  }

  /* Now copy the list to `out' (+1 for the terminating NULL). */
  out = malloc(sizeof(*out) * (size + 1));
  if (!out) {
    free_string_list(first, last);
    return NULL;
  }
  for (pos = out; size; size--, pos++) {
    struct string_list *old = first;
    *pos = first->str;
    first = first->next;
    assert(first != last || size == 1);
    free(old);
  }
  assert(first == last);
  free(last);
  *pos = NULL;

  return out;
}

/*
 * Free a NULL-terminated list of strings.
 */
static void free_usage_list(char **strings) {
  if (!strings)
    return;
  for (char **str = strings; *str; str++)
    free(*str);
  free(strings);
}

/*
 * Print the "Suboptions for [...]:" line for `opt'. Prints `prefix'
 * first, then the suboptions header, then `suffix'. `dash' is used as
 * the dash character. Returns zero on success or nonzero on error, in
 * which case `errno' will be set to indicate the error.
 */
static int print_subopts_header(const MU_OPT *opt, const char *prefix,
				const char *suffix, const char *dash,
				FORMAT_HELP_CALLBACK format,
				void *data, unsigned short *col,
				unsigned short goal,
				unsigned short width) {
  if (format(data, goal, width, col, 0, 0,
	     "%sSuboptions for ", prefix))
    return 1;

  if (opt->short_opt) {
    if (format(data, goal, width, col, 0, 0,
	       "%s%c", dash, opt->short_opt))
      return 1;
    if ((opt->long_opt || opt->env_var) &&
	format(data, goal, width, col, 0, 0, ", "))
      return 1;
  }

  if (opt->long_opt) {
    if (format(data, goal, width, col, 0, 0,
	       "%s%s%s", dash, dash, opt->long_opt))
      return 1;
    if (opt->env_var && format(data, goal, width, col, 0, 0, ", "))
      return 1;
  }

  if (opt->env_var) {
    if (format(data, goal, width, col, 0, 0, "%s", opt->env_var))
      return 1;
  }

  return format(data, goal, width, col, 0, 0, "%s", suffix);
}

/*
 * Format help for a single option. Returns zero on success or nonzero
 * on error, in which case `errno' will be set to indicate the error.
 */
static int format_opt_help(const MU_OPT *opt, enum opt_type type,
			   int has_short_opts,
			   FORMAT_HELP_CALLBACK format, void *data,
			   unsigned short goal, unsigned short width,
			   unsigned short *col, unsigned short indent,
			   unsigned short offset) {
  const char *help = get_opt_help(opt->help, type);

  /* Use 0 for the goal and width when printing the name, because we
     don't want to wrap this. */
  switch (type) {
  case OPT_TYPE_SHORT:
  case OPT_TYPE_LONG:
    if (opt->short_opt == '\0' && opt->long_opt == NULL)
      return 0;

    if (opt->short_opt) {
      if (format(data, 0, 0, col, 0, 0, "  -%c", opt->short_opt))
	return 1;
    }
    else if (has_short_opts) {
      if (format(data, 0, 0, col, 0, 0, "    "))
	return 1;
    }

    if (opt->long_opt) {
      if (format(data, 0, 0, col, 0, 0, "%c --%s",
		 opt->short_opt ? ',' : ' ', opt->long_opt))
	return 1;
    }

    break;

  case OPT_TYPE_ENV:
    if (opt->env_var == NULL)
      return 0;

    if (format(data, 0, 0, col, 0, 0, "  %s", opt->env_var))
      return 1;

    break;

  case OPT_TYPE_SUBOPT:
    if (format(data, 0, 0, col, 0, 0, "  %s", opt->subopt_name))
      return 1;

    break;
  }

  if (opt->has_arg &&
      /* Print the argument/value. We don't want to wrap this
	 either. */
      format(data, 0, 0, col, 0, 0,
	     (opt->has_arg == MU_OPT_REQUIRED) ? "%s%s" : "[%s%s]",
	     get_arg_separator(opt, type), get_arg_help(opt))) {
    return 1;
  }

  /* Put the help on a new line if there's no room on the current
     line. */
  if (get_min_indent(opt, type, has_short_opts) > indent &&
      format(data, 0, 0, col, 0, 0, "\n"))
    return 1;

  if (format(data, goal, width, col,
	     indent, indent + offset, "%s\n", help))
    return 1;

  return 0;
}

/*
 * Format help for a single option, in man(7) format. Returns zero on
 * success or nonzero on error, in which case `errno' will be set to
 * indicate the error.
 */
static int format_opt_help_man(const MU_OPT *opt, enum opt_type type,
			       FORMAT_HELP_CALLBACK format,
			       void *data, unsigned short *col) {
  char *help;
  int ret;

  /* Make sure we're actually going to print something. */
  switch (type) {
  case OPT_TYPE_SHORT:
  case OPT_TYPE_LONG:
    if (opt->short_opt == '\0' && opt->long_opt == NULL)
      return 0;
    break;
  case OPT_TYPE_ENV:
    if (opt->env_var == NULL)
      return 0;
    break;
  case OPT_TYPE_SUBOPT:
    assert(opt->subopt_name);
    break;
  }

  if (format(data, 0, 0, col, 0, 0, ".TP\n"))
    return 1;

  /* Print the option itself. */
  switch (type) {
  case OPT_TYPE_SHORT:
  case OPT_TYPE_LONG:
    if (opt->short_opt) {
      if (format(data, 0, 0, col, 0, 0,
		 "\\fB\\-%c\\fR", opt->short_opt))
	return 1;
    }

    if (opt->long_opt) {
      if (format(data, 0, 0, col, 0, 0, "%s\\fB\\-\\-%s\\fR",
		 opt->short_opt ? ", " : "", opt->long_opt))
	return 1;
    }

    break;

  case OPT_TYPE_ENV:
    if (format(data, 0, 0, col, 0, 0, "\\fB%s\\fR", opt->env_var))
      return 1;

    break;

  case OPT_TYPE_SUBOPT:
    if (format(data, 0, 0, col, 0, 0, "\\fB%s\\fR", opt->subopt_name))
      return 1;

    break;
  }

  if (opt->has_arg) {
    help = format_man_str(get_arg_help(opt), 1);
    if (!help)
      return 1;
    /* Print the argument/value. */
    ret =
      format(data, 0, 0, col, 0, 0, (opt->has_arg == MU_OPT_REQUIRED)
	     ? "%s%s" : "[%s%s]", get_arg_separator(opt, type), help);
    free(help);
    if (ret)
      return 1;
  }

  if (format(data, 0, 0, col, 0, 0, "\n"))
    return 1;

  /* Now print the help. If it's our default help, we don't want to
     format it. */
  help = opt->help ?
    format_man_str(get_opt_help(opt->help, type), 0) :
    strdup(get_opt_help(opt->help, type));
  if (!help)
    return 1;
  ret = format(data, MAN_PARAGRAPH_WIDTH, 0, col, 0, 0, "%s\n", help);
  free(help);

  return ret;
}

/*
 * Return a default help string based on `type' if `help' is NULL.
 */
static const char *get_opt_help(const char *help,
				enum opt_type type) {
  if (help)
    return help;
  switch (type) {
  case OPT_TYPE_SHORT:
  case OPT_TYPE_LONG:
    return "TODO: add help for this option";
  case OPT_TYPE_ENV:
    return "TODO: add help for this environment variable";
  case OPT_TYPE_SUBOPT:
    return "TODO: add help for this suboption";
  default:
    mu_die(-1, "Invalid option type (type code %u)\n", type);
  }
}

/*
 * Format a string for man(7) output. If the string is the `arg_help'
 * field, then you should pass `is_arg_help' as a nonzero
 * value. Otherwise, pass `is_arg_help' as zero. Returns a dynamically
 * allocated, formatted string on success or NULL on error, in which
 * case `errno' will be set to indicate the error.
 */
static char *format_man_str(const char *str, int is_arg_help) {
  size_t size;
  char *out;

  size = format_man_str_no_alloc(NULL, 0, str, is_arg_help);
  if (!size)
    return NULL;
  out = malloc(size);
  if (!out)
    return NULL;
  if (!format_man_str_no_alloc(out, size, str, is_arg_help)) {
    free(out);
    return NULL;
  }
  return out;
}

/*
 * Format a string for man(7) output. If the string is the `arg_help'
 * field, then you should pass `is_arg_help' as a nonzero
 * value. Otherwise, pass `is_arg_help' as zero. Returns the result in
 * `out', not writing more that `out_max', including the terminating
 * null byte (unless `out' is NULL, in which case no limit will be
 * enforced). If `out' is NULL, don't write anything. Returns the size
 * of `out' (or what the size would be), including the terminating
 * null byte on success or zero on error, in which case `errno' will
 * be set to indicate the error.
 *
 * Note: Currently, this is pretty hacky, and there's no real way to
 * fix that. Actually, maybe we should implement a very simple markup
 * language (such as "@var{foo}" or something like Texinfo).
 */
static size_t format_man_str_no_alloc(char *out, size_t out_max,
				      const char *str, int is_arg_help) {
  size_t index, end;
  int first;

  index = end = 0;
  for (first = 1; *str && (!out || index < out_max); first = 0) {
    if (*str == '\n') {
      int is_paragraph = 0;
      const char *p;

      /* Check if it's a new paragraph. */
      for (p = str + 1; *p && isspace(*p); p++) {
	/* Don't break when we see a newline because we want to skip
	   all the other newlines as well. */
	if (*p == '\n')
	  is_paragraph = 1;
      }

      if (is_paragraph) {
	/* Go back to the last newline. */
	do {
	  p--;
	  assert(p > str);
	} while (*p != '\n');
	p++;

	/* Start a new paragraph. */
	write_out(out, &index, out_max, "\n.PP\n");
	str = p;
      }
      else {
	/* It's not a paragraph, but put an explicit line break
	   here. */
	write_out(out, &index, out_max, "\n.br\n");
	str++;
      }
    }

    if (!end && (first || is_arg_help || !is_word_char(str[-1]))) {
      /* Check for options we should put in bold. */
      if (!is_arg_help && *str == '-') {
	for (end = 0; str[end] && is_opt_char(str[end]); end++);
	if (end && !is_word_char(str[end])) {
	  int opt = 0;

	  /* Perform one more check: make sure it actually contains a
	     letter. */
	  for (size_t i = 0; i < end; i++) {
	    if (isalpha(str[i])) {
	      opt = 1;
	      break;
	    }
	  }

	  if (opt) {
	    /* It might actually be an option; put it in bold. (Note:
	       this is still just a guess.) */
	    if (write_out(out, &index, out_max, "\\fB"))
	      return 0;
	    continue;
	  }
	}
      }

      /* Check for metasyntactic variables we should put in italics. */
      for (end = 0; str[end] && is_var_char(str[end]); end++);
      /* Don't treat "A" and "I" as metasyntactic variables unless
	 `is_arg_help', because they are English words. This is pretty
	 hacky, and they might actually be metasyntactic variables
	 after all... */
      if (end && (is_arg_help || !is_word_char(str[end])) &&
	  !(!is_arg_help && end == 1 && strchr("AI", *str))) {
	int var = 0;

	/* Perform one more check: make sure it actually contains a
	   letter. */
	for (size_t i = 0; i < end; i++) {
	  if (isalpha(str[i])) {
	    var = 1;
	    break;
	  }
	}

	if (var) {
	  /* It might actually be a metasyntactic variable; italicize
	     it. (Note: this is still just a guess.) */
	  if (write_out(out, &index, out_max, "\\fI"))
	    return 0;
	  continue;
	}
      }
    }

    /* Print the character verbatim or escape it. */
    if (write_out(out, &index, out_max, "%s%c",
		  strchr("\\-", *str) ? "\\" : "", *str))
      return 0;
    str++;
    if (end && !--end && write_out(out, &index, out_max, "\\fR"))
      return 0;
  }

  if (out)
    assert(index <= out_max);

  /* Terminate the output string. */
  if (index && index == out_max)
    index--;
  if (out && out_max)
    out[index] = '\0';
  index++;

  /* Return the size (or would be size) of the output string. */
  return index;
}

/*
 * Write to `out', at index '*index', not exceeding `out_max'. Updates
 * '*index' to the new index. Returns 0 on success or nonzero on
 * error, in which case `errno' will be set to indicate the error.
 */
static int write_out(char *out, size_t *index, size_t out_max,
		     const char *fmt, ...) {
  int len;
  char *str;
  va_list ap;

  va_start(ap, fmt);
  len = vasprintf(&str, fmt, ap);
  if (len < 0)
    return 1;
  va_end(ap);

  if (out) {
    size_t left = out_max - *index;
    if (len > left)
      len = left - 1;
    memcpy(out + *index, str, len);
  }
  *index += len;

  free(str);

  return 0;
}

/*
 * Check if `c' is a valid charcter for a metasyntactic variable.
 */
static int is_var_char(int c) {
  return isupper(c) || isdigit(c) || strchr("-_", c);
}

/*
 * Check if `c' is a valid option character.
 */
static int is_opt_char(int c) {
  return isalnum(c) || strchr("-_", c);
}

/*
 * Check if `c' is a character that could be part of a word.
 */
static int is_word_char(int c) {
  return isalnum(c);
}

/*
 * A wrapper around mu_format() to be passed as a callback to
 * format_help_type().
 */
static int format_help_file_callback(void *data, unsigned short goal,
				     unsigned short width,
				     unsigned short *col,
				     unsigned short indent,
				     unsigned short subindent,
				     const char *fmt, ...) {
  FILE *file = data;
  int ret;
  va_list ap;

  va_start(ap, fmt);
  ret = mu_vformat(file, goal, width, col,
		   indent, subindent, fmt, ap);
  va_end(ap);

  return ret;
}

/*
 * Format a string and add the result to the string list pointed to by
 * `data'. Returns zero on success or nonzero on error, in which case
 * `errno' will be set to indicate the error.
 */
static int format_help_str_callback(void *data, unsigned short goal,
				    unsigned short width,
				    unsigned short *col,
				    unsigned short indent,
				    unsigned short subindent,
				    const char *fmt, ...) {
  struct string_list **list_ptr = data, *list = *list_ptr;
  va_list ap;

  va_start(ap, fmt);
  list->str =
    mu_vformat_string(goal, width, col, indent, subindent, fmt, ap);
  va_end(ap);

  if (!list->str)
    return 1;

  list->next = malloc(sizeof(*(list->next)));
  list = list->next;
  if (!list) {
    free(list->str);
    return 1;
  }

  *list_ptr = list;
  return 0;
}

/*
 * Concatenate all strings in the string list starting at first and
 * ending at last, while also freeing the entire list, and return the
 * result. Returns NULL on error, in which case `errno' will be set to
 * indicate the error. Even on error, the entire list will be freed.
 */
static char *concat_string_list(struct string_list *first,
				struct string_list *last) {
  size_t size;
  char *str;

  /* Get the size of the strings when concatenated. */
  size = 0;
  for (struct string_list *s = first; s != last; s = s->next)
    size += strlen(s->str);
  size++;			/* For '\0'. */

  str = malloc(sizeof(*str) * size);
  if (!str) {
    free_string_list(first, last);
    return NULL;
  }

  /* Concatenate all of the strings. */
  str[0] = '\0';
  while (first != last) {
    struct string_list *old = first;

    /* This assertion is fairly expensive, but assertions can be
       turned off by defining NDEBUG, so I think we should keep it. */
    assert(strlen(str) + strlen(first->str) < size);
    strcat(str, first->str);

    first = first->next;
    free(old->str);
    free(old);
  }
  free(last);

  /* Return the result. */
  return str;
}

/*
 * Free a list of strings.
 */
static void free_string_list(struct string_list *first,
			     struct string_list *last) {
  while (first != last) {
    struct string_list *old = first;
    first = first->next;
    free(old->str);
    free(old);
  }
  free(last);
}

/*
 * Format and print a help message to `stdout' (or piped to man). Does
 * not return on success.
 */
static int help_option_callback(int has_arg, const char *arg,
				void *data, char *err) {
  struct help_callback_data *d = data;

  if (!has_arg) {
    arg = d->default_format;
    has_arg = !!arg;
  }

  if (has_arg) {
    size_t arg_len = strlen(arg);

    /* Get the help format. */
    if (!strncasecmp(arg, "man", arg_len))
      return man_option_callback(data, err);

    if (strncasecmp(arg, "plain", arg_len) &&
	strncasecmp(arg, "text", arg_len)) {
      snprintf(err, MU_OPT_ERR_MAX, "invalid help format: %s", arg);
      return 1;
    }
  }

  if (mu_format_help(stdout, d->prog_name, d->usage,
		     d->desc, d->notes, d->options) ||
      fflush(stdout) == EOF) {
    snprintf(err, MU_OPT_ERR_MAX,
	     "cannot print help message: %s", strerror(errno));
    return 1;
  }

  /* We don't really need to do this since we're about to exit(), but
     it makes valgrind happy. */
  mu_free_opts(d->options);

  exit(0);
}

/*
 * Format and print a help message in man(7) format. Does not return
 * on success.
 */
static int man_option_callback(void *data, char *err) {
  struct help_callback_data *d = data;
  FILE *output = stdout;
#ifdef USE_MAN_PIPE
  int do_man_pipe = 0;
  pid_t man_pid;
  int wstatus;
  int man_pipe[2], err_pipe[2];
#endif

#ifdef USE_MAN_PIPE
  /* Check whether we should pipe the help output to `man'. */
  do_man_pipe = isatty(STDOUT_FILENO);
  if (do_man_pipe && (pipe(man_pipe) || pipe(err_pipe))) {
    /* An error occured while setting up the pipe. Print a warning and
       don't pipe to man after all. */
    fprintf(stderr, "%s: warning: cannot set up pipe: %s\n",
	    d->prog_name, strerror(errno));
    do_man_pipe = 0;
  }

  if (do_man_pipe) {
    man_pid = fork();
    if (man_pid < 0) {
      /* An error occured. Print a warning and don't pipe to man after
	 all. */
      fprintf(stderr, "%s: warning: cannot fork: %s\n",
	      d->prog_name, strerror(errno));
      do_man_pipe = 0;
    }
  }

  if (do_man_pipe) {
    char *chld_err;

    if (!man_pid) {
      /* We are the child. We don't need the write end of `man_pipe'
	 or the read end of `err_pipe'. */
      if (close(man_pipe[1]) || close(err_pipe[0]))
	report_err(err_pipe[1], "cannot close pipes: %s",
		   strerror(errno));

      /* Redirect standard input from `man_pipe'. */
      if (dup2(man_pipe[0], STDIN_FILENO) < 0)
	report_err(err_pipe[1],
		   "cannot redirect standard input: %s",
		   strerror(errno));

      /* Make sure to close the output `err_pipe' after a successful
	 execve() so that get_err() doesn't hang in the parent. */
      if (fcntl(err_pipe[1], F_SETFD, FD_CLOEXEC) < 0)
	report_err(err_pipe[1],
		   "cannot set FD_CLOEXEC flag on error pipe (%d): %s",
		   err_pipe[1], strerror(errno));

      /* Execute `man' in such a way that it reads from standard
	 input. */
      execlp("man", "man", "-l", "-", NULL);
      report_err(err_pipe[1], "cannot execute 'man': %s",
		 strerror(errno));
    }

    /* We are the parent. We don't need the read end of `man_pipe'
       or the write end of `err_pipe'. */
    if (close(man_pipe[0]) || close(err_pipe[1])) {
      snprintf(err, MU_OPT_ERR_MAX, "cannot close pipes: %s",
	       strerror(errno));
      goto kill;
    }

    /* Check if an error occured. */
    chld_err = get_err(err_pipe[0]);

    /* We can close the error pipe now. */
    if (close(err_pipe[0])) {
      snprintf(err, MU_OPT_ERR_MAX,
	       "cannot close error pipe: %s", strerror(errno));
      return 1;
    }

    if (chld_err) {
      /* An error occured. Print a warning message. */
      fprintf(stderr, "%s: warning: %s\n", d->prog_name, chld_err);
      free(chld_err);

      /* Wait for the child... */
      do {
	pid_t pid = waitpid(man_pid, &wstatus, 0);
	if (pid < 0) {
	  snprintf(err, MU_OPT_ERR_MAX,
		   "cannot wait for child: %s", strerror(errno));
	  return 1;
	}
	assert(pid == man_pid);
      } while (!(WIFEXITED(wstatus) || WIFSIGNALED(wstatus)));

      /* And don't pipe through man after all. */
      do_man_pipe = 0;
    }
  }

  if (do_man_pipe) {
    /* Associate 'man_pipe[1]' with a stream to pass to
       mu_format_help(). */
    output = fdopen(man_pipe[1], "w");
    if (!output) {
      snprintf(err, MU_OPT_ERR_MAX, "cannot associate "
	       "file descriptor %d with a stream: %s",
	       man_pipe[1], strerror(errno));
      goto kill;
    }
  }
#endif /* USE_MAN_PIPE */

  /* Format the help text. */
  if (mu_format_help_man(output, d->prog_name, d->usage,
			 d->short_desc, d->desc, d->notes,
			 d->section, d->sect_name,
			 d->pkg_str, d->date, d->options) ||
      fflush(output) == EOF) {
    snprintf(err, MU_OPT_ERR_MAX,
	     "cannot print help message: %s", strerror(errno));
#ifdef USE_MAN_PIPE
    if (do_man_pipe) {
      fclose(output);
      close(err_pipe[0]);
      goto kill;
    }
#endif
    return 1;
  }

#ifdef USE_MAN_PIPE
  if (do_man_pipe) {
    /* Close the man pipe (closing `output' automatically closes
       'man_pipe[1]') */
    if (fclose(output)) {
      snprintf(err, MU_OPT_ERR_MAX,
	       "cannot close man pipe: %s", strerror(errno));
      close(err_pipe[0]);
      goto kill;
    }

    /* Wait for the child process. */
    do {
      pid_t pid = waitpid(man_pid, &wstatus, 0);
      if (pid < 0) {
	snprintf(err, MU_OPT_ERR_MAX,
		 "cannot wait for man: %s", strerror(errno));
	return 1;
      }
      assert(pid == man_pid);
    } while (!(WIFEXITED(wstatus) || WIFSIGNALED(wstatus)));

    if (WIFEXITED(wstatus)) {
      int status = WEXITSTATUS(wstatus);
      if (status) {
	snprintf(err, MU_OPT_ERR_MAX,
		 "man exited with status %d", status);
	return 1;
      }
    }
    else if (WIFSIGNALED(wstatus)) {
      int signal, coredump;
      char signame[SIG2STR_MAX];

      signal = WTERMSIG(wstatus);
      /* WCOREDUMP() is not specified in POSIX.1-2001 (see
	 wait(2)). */
# ifdef WCOREDUMP
      coredump = WCOREDUMP(wstatus);
# else
      coredump = 0;
# endif

      /* Get the signal name. */
      if (sig2str(signal, signame))
	mu_die(-1, "invalid signal: %d", signal);

      snprintf(err, MU_OPT_ERR_MAX,
	       /* This looks like a headline (heh heh). */
	       "man killed by SIG%s (%d)%s",
	       signame, signal, coredump ? "  (core dumped)" : "");

      return 1;
    }
    else
      mu_die(-1, "child wasn't signaled or exited");
  }
#endif /* USE_MAN_PIPE */

  /* We don't really need to do this since we're about to exit(), but
     it makes valgrind happy. */
  mu_free_opts(d->options);

  exit(0);

#ifdef USE_MAN_PIPE
 kill: {
    /* A single millisecond. */
    const struct timespec ms = {0, 1000000};
    int dead = 0;
    pid_t pid;

    /* First try to nicely ask the child to terminate. If the kill()
       fails, then it's because the child is already dead (or rather,
       undead), so we should still be able to wait for it. */
    kill(man_pid, SIGTERM);

    /* Now try and wait for the child. Give it 10 ms to die, checking
       every millisecond. */
    for (int i = 0; i < 10; i++) {
      pid = waitpid(man_pid, NULL, WNOHANG);
      if (pid) {
	/* No errors should occur. */
	assert(pid == man_pid);
	dead = 1;
	break;
      }

      /* No need for error checking; it's not worth it. */
      nanosleep(&ms, NULL);
    }

    if (!dead) {
      /* The child is still alive. Remedy that. */
      kill(man_pid, SIGKILL);
      pid = waitpid(man_pid, NULL, WNOHANG);
      /* We should definitely be able to wait for the child now. */
      assert(pid == man_pid);
    }

    return 1;
  }
#endif /* USE_MAN_PIPE */
}

/*
 * Decrement `data''s `ref_count' and free it if it reaches 0.
 */
static void free_help_callback_data(void *data) {
  struct help_callback_data *d = data;
  if (!--(d->ref_count))
    free(d);
}

#ifdef USE_MAN_PIPE
/*
 * Report an error to the parent process. Does not return.
 */
static void report_err(int fd, const char *fmt, ...) {
  char *err;
  unsigned err_len;
  va_list ap;
  int success;

  va_start(ap, fmt);
  err_len = mu_xvasprintf(&err, fmt, ap);
  va_end(ap);

  /* Send the length of the error message first, and the message
     itself second. */
  success =
    write(fd, &err_len,	sizeof(err_len)) == sizeof(err_len) &&
    write(fd, err, sizeof(*err) * err_len) == sizeof(*err) * err_len;
  if (!success) {
    /* We can't write the error message, so just print it instead. */
    fputs(err, stderr);
    putc('\n', stderr);
  }
  free(err);

  exit(-1);
}

/*
 * Get an error from the child process. Returns a dynamically
 * allocated error string if there was en error, or NULL if there
 * wasn't.
 */
static char *get_err(int fd) {
  ssize_t bytes_read;
  char *err;
  unsigned err_len;

  /* First get the length of the error string. */
  bytes_read = read(fd, &err_len, sizeof(err_len));
  if (!bytes_read)
    return NULL;
  if (bytes_read < 0) {
    mu_xasprintf(&err, "cannot get error string: %s",
		 strerror(errno));
    return err;
  }
  if (bytes_read != sizeof(err_len))
    return mu_xstrdup("cannot get error string");

  /* Now get the actual error string (+1 for '\0'). */
  err = mu_xmalloc(sizeof(*err) * (err_len + 1));
  bytes_read = read(fd, err, sizeof(*err) * err_len);
  if (bytes_read < 0) {
    free(err);
    mu_xasprintf(&err, "cannot get error string: %s",
		 strerror(errno));
    return err;
  }
  /* If we couldn't read the whole string, I guess just return what we
     got. */
  assert(bytes_read <= err_len);
  err_len = bytes_read;

  /* Terminate the error string. */
  err[err_len] = '\0';

  return err;
}
#endif /* USE_MAN_PIPE */
