/********************************************************************\
 * Mu (Miscellaneous Utilities) is a general convenience library.   *
 * Copyright (C) 2020  Asher Gordon <AsDaGo@posteo.net>             *
 *                                                                  *
 * This file is part of Mu.                                         *
 *                                                                  *
 * Mu is free software: you can redistribute it and/or modify it    *
 * under the terms of the GNU General Public License as published   *
 * by the Free Software Foundation, either version 3 of the         *
 * License, or (at your option) any later version.                  *
 *                                                                  *
 * Mu is distributed in the hope that it will be useful, but        *
 * WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU General Public License for more details.                     *
 *                                                                  *
 * You should have received a copy of the GNU General Public        *
 * License along with Mu.  If not, see                              *
 * <https://www.gnu.org/licenses/>.                                 *
\********************************************************************/

/*
 * options.c -- check option parsing
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <assert.h>

/* Mu includes */
#include <options.h>
#include <safe.h>

/* The name of this test, what it tests, a description of the command
   line options and arguments, and a short usage description. */
#define TEST_NAME	"Options"
#define TEST_SUBJECT	"option parsing"
#define TEST_USAGE						     \
  "[PERMUTE|BUNDLE|ALLOW-INVALID]... [ENV_VAR=VAL]... [OPTIONS]..."
#define TEST_DESC						\
  "The first arguments specified are the flags to pass to "	\
  "mu_parse_opts(). The arguments after (and including) the "	\
  "first non-flag argument are options and positional "		\
  "arguments (which will be printed)."

#include "test.h"

extern char **environ;

#define FILE_OPT						\
  .has_arg	= MU_OPT_REQUIRED,				\
  .arg_type	= MU_OPT_FILE,					\
  .file_mode	= "r",						\
  .argstr	= &file_name,					\
  .arg		= &file,					\
  .callback_file = file_callback,				\
  .arg_help	= "FILE",					\
  .help		= "read FILE and print some of its contents"
#define DIR_OPT						\
  .has_arg	= MU_OPT_REQUIRED,			\
  .arg_type	= MU_OPT_DIRECTORY,			\
  .arg		= &dir,					\
  .callback_directory = dir_callback,			\
  .arg_help	= "DIR",				\
  .help		= "list the contents of DIR"

/*
 * Static helper functions
 */
static int file_callback(int, const char *, FILE *, void *, char *);
static int dir_callback(int, const char *, DIR *, void *, char *);
static int str_callback(int, const char *, void *, char *);
static int int_callback(int, long, void *, char *);
static int float_callback(int, double, void *, char *);
static int bool_callback(int, int, void *, char *);
static int error_callback(void *, char *);
static int print_opt(void *, char *);
static int print_opt_arg(int, const char *, void *, char *);
static int subopt_no_arg(void *, char *);
static int subopt_arg(int, const char *, void *, char *);
static int read_and_print(FILE *);

int main(int argc, char **argv) {
  const char *file_name;
  FILE *file;
  DIR *dir;
  int found_subopt, found_subopt_arg;
  MU_OPT suboptions[] = {
    {
     .subopt_name	= "file",
     FILE_OPT
    },
    {
     .subopt_name	= "directory",
     DIR_OPT
    },
    {
     .subopt_name	= "none",
     .env_var		= "SUBOPT_NONE",
     .has_arg		= MU_OPT_NONE,
     .callback_none	= subopt_no_arg,
     .help		= "a suboption taking no argument"
    },
    {
     .subopt_name	= "optional",
     .env_var		= "SUBOPT_OPTIONAL",
     .has_arg		= MU_OPT_OPTIONAL,
     .arg_type		= MU_OPT_STRING,
     .callback_string	= subopt_arg,
     .cb_data		= (void *)0,
     .help		= "a suboption taking an optional argument",
    },
    {
     .subopt_name	= "required",
     .env_var		= "SUBOPT_REQUIRED",
     .has_arg		= MU_OPT_REQUIRED,
     .arg_type		= MU_OPT_STRING,
     .callback_string	= subopt_arg,
     .cb_data		= (void *)1,
     .help		= "a suboption taking a required argument",
    },
    { 0 }
  };
  MU_OPT options[] = {
    {
     .short_opt	= 'f',
     .long_opt	= "file",
     .env_var	= "FILE",
     FILE_OPT
    },
    {
     .short_opt	= 'd',
     .long_opt	= "directory",
     .env_var	= "DIRECTORY",
     DIR_OPT
    },
    {
     .short_opt	= 's',
     .long_opt	= "subopts",
     .env_var	= "SUBOPTS",
     .has_arg	= MU_OPT_OPTIONAL,
     .arg_type	= MU_OPT_SUBOPT,
     .found_opt	= &found_subopt,
     .found_arg	= &found_subopt_arg,
     .subopts	= suboptions,
     .arg_help	= "SUBOPTS",
     .help	= "parse SUBOPTS as suboptions"
    },
    {
     .short_opt	= 'a',
     .long_opt	= "string",
     .env_var	= "STRING",
     .has_arg	= MU_OPT_REQUIRED,
     .arg_type	= MU_OPT_STRING,
     .callback_string = str_callback,
     .help	= "take a string argument"
    },
    {
     .short_opt	= 'i',
     .long_opt	= "integer",
     .env_var	= "INTEGER",
     .has_arg	= MU_OPT_REQUIRED,
     .arg_type	= MU_OPT_INT,
     .ibound	= { -10, 10 },
     .callback_int = int_callback,
     .help	=
     "take an integer argument between -10 and 10 inclusive"
    },
    {
     .short_opt	= 'n',
     .long_opt	= "float",
     .env_var	= "FLOAT",
     .has_arg	= MU_OPT_REQUIRED,
     .arg_type	= MU_OPT_FLOAT,
     .callback_float = float_callback,
     .fbound	= { -10, 10 },
     .help	=
     "take a floating-point number argument between -10 and 10 "
     "inclusive"
    },
    {
     .short_opt	= 'b',
     .long_opt	= "boolean",
     .env_var	= "BOOLEAN",
     .has_arg	= MU_OPT_REQUIRED,
     .arg_type	= MU_OPT_BOOL,
     .callback_bool = bool_callback,
     .help	= "take a boolean argument"
    },
    {
     .short_opt	= 'e',
     .long_opt	= "error",
     .env_var	= "ERROR",
     .has_arg	= MU_OPT_NONE,
     .callback_none = error_callback,
     .help	= "cause fatal error"
    },
    {
     .long_opt	= "foobarbaz",
     .has_arg	= MU_OPT_OPTIONAL,
     .arg_type	= MU_OPT_STRING,
     .callback_string = print_opt_arg,
     .cb_data	= "foobarbaz",
     .help	= "an option named 'foobarbaz'"
    },
    {
     .long_opt	= "foobazbar",
     .has_arg	= MU_OPT_REQUIRED,
     .arg_type	= MU_OPT_STRING,
     .callback_string = print_opt_arg,
     .cb_data	= "foobazbar",
     .help	= "an option named 'foobazbar'"
    },
    {
     .long_opt	= "foobarbazqux",
     .has_arg	= MU_OPT_OPTIONAL,
     .arg_type	= MU_OPT_STRING,
     .callback_string = print_opt_arg,
     .cb_data	= "foobarbazqux",
     .help	= "an option named 'foobarbazqux'"
    },
    {
     .long_opt	= "vroom",
     .has_arg	= MU_OPT_OPTIONAL,
     .arg_type	= MU_OPT_STRING,
     .callback_string = print_opt_arg,
     .cb_data	= "vroom",
     .help	=
     "a long option that doesn't share a first letter with a "
     "short option (that's right, all the others were taken)"
    },
    {
     .short_opt	= 'w',
     .has_arg	= MU_OPT_NONE,
     .callback_none = print_opt,
     .cb_data	= "w",
     .help	= "an option named 'w'"
    },
    {
     .short_opt	= 'x',
     .has_arg	= MU_OPT_NONE,
     .callback_none = print_opt,
     .cb_data	= "x",
     .help	= "an option named 'x'"
    },
    {
     .short_opt	= 'y',
     .has_arg	= MU_OPT_OPTIONAL,
     .arg_type	= MU_OPT_STRING,
     .callback_string = print_opt_arg,
     .cb_data	= "y",
     .help	= "an option named 'y'"
    },
    {
     .short_opt	= 'z',
     .has_arg	= MU_OPT_REQUIRED,
     .arg_type	= MU_OPT_STRING,
     .callback_string = print_opt_arg,
     .cb_data	= "z",
     .help	= "an option named 'z'"
    },
    TEST_OPT_END
  };
  int ret;
  int flags = 0;

  test_create_help_option(options);

  /* Get the flags to use for parsing options and also set any
     environment variables passed on the command line. */
  while (argc > 1) {
    char *arg = argv[1];

    if (!strcmp(arg, "PERMUTE"))
      flags |= MU_OPT_PERMUTE;
    else if (!strcmp(arg, "BUNDLE"))
      flags |= MU_OPT_BUNDLE;
    else if (!strcmp(arg, "ALLOW-INVALID"))
      flags |= MU_OPT_ALLOW_INVALID;
    else if (*arg != '-' && strchr(arg, '=')) {
      if (putenv(arg))
	mu_die(1, "%s: cannot set environment variable '%s': %s\n",
	       argv[0], arg, strerror(errno));
    }
    else
      break;

    MU_SHIFT_ARGS(1);
  }

  /* Now parse the options. */
  ret = mu_parse_opts(argc, argv, environ, options, flags);
  if (MU_OPT_ERR(ret))
    return 10;		    /* 10 to indicate failed option parsing */
  MU_SHIFT_ARGS(ret);

  if (file) {
    printf("main: reading from %s\n", file_name);
    if (read_and_print(file))
      mu_die(1, "%s: cannot read a line from %s and print to "
	     "stdout: %s\n", argv[0], file_name, strerror(errno));
    if (fclose(file))
      mu_die(1, "%s: cannot close %s: %s\n",
	     argv[0], file_name, strerror(errno));
  }

  if (found_subopt)
    printf("Found the --subopt option, %s.\n", found_subopt_arg ?
	   "and it had an argument" : "but it had no arguments");

  if (argc > 1) {
    /* Print the remaining arguments. */
    printf("%d remaining argument%s: {",
	   argc - 1, (argc - 1 == 1) ? "" : "s");
    for (int i = 1; i < argc; i++)
      printf("%s'%s'", (i > 1) ? ", " : "", argv[i]);
    puts("}");
  }

  return 0;
}

/***************************\
|* Static helper functions *|
\***************************/

/*
 * Read a file and print one line of its contents.
 */
static int file_callback(int has_arg, const char *argstr, FILE *arg,
			 void *data, char *err) {
  printf("file_callback: reading from %s\n", argstr);
  if (read_and_print(arg)) {
    snprintf(err, MU_OPT_ERR_MAX,
	     "cannot read a line from %s and print to stdout: %s",
	     argstr, strerror(errno));
    return 1;
  }
  return 0;
}

/*
 * List a directory's contents.
 */
static int dir_callback(int has_arg, const char *argstr, DIR *arg,
			void *data, char *err) {
  struct dirent **files;
  size_t n_files;

  printf("Reading '%s'...\n", argstr);

  /* Get the number of files. */
  errno = 0;
  for (n_files = 0; readdir(arg); n_files++)
    errno = 0;
  if (errno)
    goto error;
  rewinddir(arg);

  /* Now get the actual files. */
  files = malloc(sizeof(*files) * n_files);
  if (!files)
    goto error;
  for (size_t i = 0; i < n_files; i++) {
    errno = 0;
    files[i] = readdir(arg);
    if (!files[i]) {
      free(files);
      goto error;
    }
  }
  assert(!readdir(arg) && !errno);

  /* Sort them. */
  qsort(files, n_files, sizeof(*files),
	(int (*)(const void *, const void *))alphasort);

  /* Print them. */
  for (size_t i = 0; i < n_files; i++)
    puts(files[i]->d_name);

  free(files);

  return 0;

 error:
  snprintf(err, MU_OPT_ERR_MAX, "cannot read directory %s: %s",
	   argstr, strerror(errno));
  return 1;
}

/*
 * Print the string argument to standard output.
 */
static int str_callback(int has_arg, const char *arg,
			void *data, char *err) {
  if (printf("Found a string argument: '%s'\n", arg) < 0) {
    snprintf(err, MU_OPT_ERR_MAX,
	     "cannot write to standard output: %s", strerror(errno));
    return 1;
  }
  return 0;
}

/*
 * Print the integer argument to standard output.
 */
static int int_callback(int has_arg, long arg,
			void *data, char *err) {
  if (printf("Found an integer argument: %ld\n", arg) < 0) {
    snprintf(err, MU_OPT_ERR_MAX,
	     "cannot write to standard output: %s", strerror(errno));
    return 1;
  }
  return 0;
}

/*
 * Print the floating-point number argument to standard output.
 */
static int float_callback(int has_arg, double arg,
			  void *data, char *err) {
  if (printf("Found a floating-point number argument: %g\n", arg)
      < 0) {
    snprintf(err, MU_OPT_ERR_MAX,
	     "cannot write to standard output: %s", strerror(errno));
    return 1;
  }
  return 0;
}

/*
 * Print the boolean argument to standard output.
 */
static int bool_callback(int has_arg, int arg,
			 void *data, char *err) {
  if (printf("Found a boolean argument: %s\n",
	     arg ? "true" : "false") < 0) {
    snprintf(err, MU_OPT_ERR_MAX,
	     "cannot write to standard output: %s", strerror(errno));
    return 1;
  }
  return 0;
}

/*
 * Return an error.
 */
static int error_callback(void *data, char *err) {
  strncpy(err, "an error occurred", MU_OPT_ERR_MAX);
  return 1;
}

/*
 * Print a message when an option that doesn't take an argument is
 * found.
 */
static int print_opt(void *data, char *err) {
  const char *name = data;
  if (printf("Found option: '%s'\n", name) < 0) {
    snprintf(err, MU_OPT_ERR_MAX, "cannot print option '%s': %s",
	     name, strerror(errno));
    return 1;
  }
  return 0;
}

/*
 * Print a message when an option that takes an argument is found.
 */
static int print_opt_arg(int has_arg, const char *arg,
			 void *data, char *err) {
  const char *name = data;
  if (printf("Found option: '%s', ", name) < 0 ||
      (has_arg ?
       (printf("with an argument: '%s'.\n", arg) < 0) :
       (puts("without an argument.") == EOF))) {
    snprintf(err, MU_OPT_ERR_MAX,
	     "cannot print option '%s' (%s an argument%s%s%s): %s",
	     name, has_arg ? "with" : "without", has_arg ? ": '" : "",
	     has_arg ? arg : "", has_arg ? "'" : "", strerror(errno));
    return 1;
  }
  return 0;
}

/*
 * Print a message when we receive a suboption without an argument.
 */
static int subopt_no_arg(void *data, char *err) {
  if (puts("Found a suboption which doesn't take an argument.")
      == EOF) {
    snprintf(err, MU_OPT_ERR_MAX,
	     "cannot write to standard output: %s", strerror(errno));
    return 1;
  }
  return 0;
}

/*
 * Print a message when we receive a suboption with an argument (data
 * is "optional" for an optional argument, or "required" for a
 * required argument).
 */
static int subopt_arg(int has_arg, const char *arg,
		      void *data, char *err) {
  int required = !!data;
  if (printf("Found a suboption which takes %s argument, %s",
	     required ? "a required" : "an optional",
	     has_arg ? "and the argument is " :
	     "but it did not have an argument") < 0 ||
      (has_arg && printf("'%s'", arg) < 0) ||
      puts(".") == EOF) {
    snprintf(err, MU_OPT_ERR_MAX,
	     "cannot write to standard output: %s", strerror(errno));
    return 1;
  }
  return 0;
}

/*
 * Read a single line from `stream' and print it to standard
 * output. Return zero on success or nonzero on error, in which case
 * `errno' will be set to indicate the error.
 */
static int read_and_print(FILE *stream) {
  int c;
  while ((c = getc(stream)) != EOF && c != '\n') {
    if (putchar(c) == EOF)
      return 1;
  }
  return ferror(stream) || putchar('\n') == EOF || fflush(stdout);
}
